from typing import Any, Callable

import dask.dataframe as dd
import ibis
from ibis.backends.dask import Backend
from triad.utils.assertion import assert_or_throw

from fugue import DataFrame, DataFrames, ExecutionEngine
from fugue_dask.dataframe import DaskDataFrame
from fugue_dask.execution_engine import DaskExecutionEngine
from fugue_ibis import IbisTable
from fugue_ibis._utils import to_ibis_schema, to_schema
from fugue_ibis.execution.ibis_engine import IbisEngine, parse_ibis_engine


class DaskIbisEngine(IbisEngine):
    def __init__(self, execution_engine: ExecutionEngine) -> None:
        assert_or_throw(
            isinstance(execution_engine, DaskExecutionEngine),
            lambda: ValueError(
                f"DaskIbisEngine must use DaskExecutionEngine ({execution_engine})"
            ),
        )
        super().__init__(execution_engine)

    def select(
        self, dfs: DataFrames, ibis_func: Callable[[ibis.BaseBackend], IbisTable]
    ) -> DataFrame:
        pdfs = {
            k: self.execution_engine.to_df(v).native  # type: ignore
            for k, v in dfs.items()
        }
        be = _BackendWrapper().connect(pdfs)
        be.set_schemas(dfs)
        expr = ibis_func(be)
        schema = to_schema(expr.schema())
        result = expr.compile()
        assert_or_throw(
            isinstance(result, dd.DataFrame),
            lambda: ValueError(f"result must be a Dask DataFrame ({type(result)})"),
        )
        return DaskDataFrame(result, schema=schema)


@parse_ibis_engine.candidate(
    lambda obj, *args, **kwargs: isinstance(obj, DaskExecutionEngine)
)
def _to_dask_ibis_engine(obj: Any, engine: ExecutionEngine) -> IbisEngine:
    return DaskIbisEngine(engine)


class _BackendWrapper(Backend):
    def set_schemas(self, dfs: DataFrames) -> None:
        self._schemas = {k: to_ibis_schema(v.schema) for k, v in dfs.items()}

    def table(self, name: str, schema: Any = None):
        return super().table(
            name,
            schema=self._schemas[name]
            if schema is None and name in self._schemas
            else schema,
        )



import itertools
import logging
from abc import abstractmethod
from typing import Any, Callable, Dict, List, Optional, Type

import ibis
from ibis import BaseBackend
from triad import assert_or_throw

from fugue import StructuredRawSQL
from fugue.bag import Bag, LocalBag
from fugue.collections.partition import (
    BagPartitionCursor,
    PartitionCursor,
    PartitionSpec,
    parse_presort_exp,
)
from fugue.dataframe import DataFrame, DataFrames, LocalDataFrame
from fugue.dataframe.utils import get_join_schemas
from fugue.execution.execution_engine import ExecutionEngine, MapEngine, SQLEngine

from ._compat import IbisTable
from ._utils import to_ibis_schema
from .dataframe import IbisDataFrame

_JOIN_RIGHT_SUFFIX = "_ibis_y__"
_GEN_TABLE_NAMES = (f"_fugue_temp_table_{i:d}" for i in itertools.count())


class IbisSQLEngine(SQLEngine):
    """Ibis SQL backend base implementation.

    :param execution_engine: the execution engine this sql engine will run on
    """

    @property
    @abstractmethod
    def backend(self) -> BaseBackend:  # pragma: no cover
        raise NotImplementedError

    def select(self, dfs: DataFrames, statement: StructuredRawSQL) -> DataFrame:
        return self.to_df(
            self.query_to_table(
                statement.construct(dialect=self.dialect, log=self.log),
                dfs,
            )
        )

    def query_to_table(self, statement: str, dfs: Dict[str, Any]) -> IbisTable:
        cte: List[str] = []
        for k, v in dfs.items():
            idf = self.to_df(v)
            cte.append(k + " AS (" + idf.to_sql() + ")")  # type: ignore
        if len(cte) > 0:
            sql = "WITH " + ",\n".join(cte) + "\n" + statement
        else:
            sql = statement
        return self.backend.sql(sql)

    @abstractmethod
    def encode_column_name(self, name: str) -> str:  # pragma: no cover
        raise NotImplementedError

    def get_temp_table_name(self) -> str:
        return next(_GEN_TABLE_NAMES)

    @abstractmethod
    def persist(
        self, df: DataFrame, lazy: bool = False, **kwargs: Any
    ) -> DataFrame:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    def sample(
        self,
        df: DataFrame,
        n: Optional[int] = None,
        frac: Optional[float] = None,
        replace: bool = False,
        seed: Optional[int] = None,
    ) -> DataFrame:  # pragma: no cover
        raise NotImplementedError

    def join(
        self,
        df1: DataFrame,
        df2: DataFrame,
        how: str,
        on: Optional[List[str]] = None,
    ) -> DataFrame:
        _df1 = self.to_df(df1)
        _df2 = self.to_df(df2)
        key_schema, end_schema = get_join_schemas(_df1, _df2, how=how, on=on)
        on_fields = [_df1.native[k] == _df2.native[k] for k in key_schema]
        if ibis.__version__ < "6":
            suffixes: Dict[str, Any] = dict(suffixes=("", _JOIN_RIGHT_SUFFIX))
        else:  # pragma: no cover
            # breaking change in ibis 6.0
            suffixes = dict(lname="", rname=_JOIN_RIGHT_SUFFIX)
        if how.lower() == "cross":
            tb = _df1.native.cross_join(_df2.native, **suffixes)
        elif how.lower() == "right_outer":
            tb = _df2.native.left_join(_df1.native, on_fields, **suffixes)
        elif how.lower() == "left_outer":
            tb = _df1.native.left_join(_df2.native, on_fields, **suffixes)
        elif how.lower() == "full_outer":
            tb = _df1.native.outer_join(_df2.native, on_fields, **suffixes)
            cols: List[Any] = []
            for k in end_schema.names:
                if k not in key_schema:
                    cols.append(k)
                else:
                    cols.append(
                        ibis.coalesce(tb[k], tb[k + _JOIN_RIGHT_SUFFIX]).name(k)
                    )
            tb = tb[cols]
        elif how.lower() in ["semi", "left_semi"]:
            tb = _df1.native.semi_join(_df2.native, on_fields, **suffixes)
        elif how.lower() in ["anti", "left_anti"]:
            tb = _df1.native.anti_join(_df2.native, on_fields, **suffixes)
        else:
            tb = _df1.native.inner_join(_df2.native, on_fields, **suffixes)
        return self.to_df(tb[end_schema.names], schema=end_schema)

    def union(self, df1: DataFrame, df2: DataFrame, distinct: bool = True) -> DataFrame:
        _df1 = self.to_df(df1)
        _df2 = self.to_df(df2)
        tb = _df1.native.union(_df2.native, distinct=distinct)
        return self.to_df(tb, df1.schema)

    def subtract(
        self, df1: DataFrame, df2: DataFrame, distinct: bool = True
    ) -> DataFrame:
        _df1 = self.to_df(df1)
        _df2 = self.to_df(df2)
        tb = _df1.native.difference(_df2.native, distinct=distinct)
        return self.to_df(tb, df1.schema)

    def intersect(
        self, df1: DataFrame, df2: DataFrame, distinct: bool = True
    ) -> DataFrame:
        _df1 = self.to_df(df1)
        _df2 = self.to_df(df2)
        tb = _df1.native.intersect(_df2.native, distinct=distinct)
        return self.to_df(tb, df1.schema)

    def distinct(self, df: DataFrame) -> DataFrame:
        _df = self.to_df(df)
        tb = _df.native.distinct()
        return self.to_df(tb, df.schema)

    def dropna(
        self,
        df: DataFrame,
        how: str = "any",
        thresh: int = None,
        subset: Optional[List[str]] = None,
    ) -> DataFrame:
        schema = df.schema
        if subset is not None:
            schema = schema.extract(subset)
        _df = self.to_df(df)
        if thresh is None:
            tb = _df.native.dropna(subset=subset, how=how)
            return self.to_df(tb, df.schema)
        assert_or_throw(
            how == "any", ValueError("when thresh is set, how must be 'any'")
        )
        sm = None
        for col in schema.names:
            expr = _df.native[col].isnull().ifelse(0, 1)
            if sm is None:
                sm = expr
            else:
                sm = sm + expr
        tb = _df.native.filter(sm >= ibis.literal(thresh))
        return self.to_df(tb, df.schema)

    def fillna(self, df: DataFrame, value: Any, subset: List[str] = None) -> DataFrame:
        def _build_value_dict(names: List[str]) -> Dict[str, str]:
            if not isinstance(value, dict):
                return {n: value for n in names}
            else:
                return {n: value[n] for n in names}

        names = df.columns
        if isinstance(value, dict):
            # subset should be ignored
            names = list(value.keys())
        elif subset is not None:
            st = set(names)
            assert_or_throw(
                st.issuperset(subset),
                ValueError(f"{names} is not a superset of {subset}"),
            )
            names = subset
        vd = _build_value_dict(names)
        assert_or_throw(
            all(v is not None for v in vd.values()),
            ValueError("fillna value can not be None or contain None"),
        )
        tb = self.to_df(df).native
        cols = [
            ibis.coalesce(tb[f], ibis.literal(vd[f])).name(f) if f in names else tb[f]
            for f in df.columns
        ]
        return self.to_df(tb[cols], schema=df.schema)

    def take(
        self,
        df: DataFrame,
        n: int,
        presort: str,
        na_position: str = "last",
        partition_spec: Optional[PartitionSpec] = None,
    ) -> DataFrame:
        partition_spec = partition_spec or PartitionSpec()
        assert_or_throw(
            isinstance(n, int),
            ValueError("n needs to be an integer"),
        )

        if presort is not None and presort != "":
            _presort = parse_presort_exp(presort)
        else:
            _presort = partition_spec.presort
        tbn = "_temp"
        idf = self.to_df(df)

        if len(_presort) == 0:
            if len(partition_spec.partition_by) == 0:
                return self.to_df(idf.native.head(n), schema=df.schema)
            pcols = ", ".join(
                self.encode_column_name(x) for x in partition_spec.partition_by
            )
            sql = (
                f"SELECT * FROM ("
                f"SELECT *, ROW_NUMBER() OVER (PARTITION BY {pcols}) "
                f"AS __fugue_take_param FROM {tbn}"
                f") WHERE __fugue_take_param<={n}"
            )
            tb = self.query_to_table(sql, {tbn: idf})
            return self.to_df(tb[df.columns], schema=df.schema)

        sorts: List[str] = []
        for k, v in _presort.items():
            s = self.encode_column_name(k)
            s += " ASC" if v else " DESC"
            s += " NULLS FIRST" if na_position == "first" else " NULLS LAST"
            sorts.append(s)
        sort_expr = "ORDER BY " + ", ".join(sorts)

        if len(partition_spec.partition_by) == 0:
            sql = f"SELECT * FROM {tbn} {sort_expr} LIMIT {n}"
            tb = self.query_to_table(sql, {tbn: idf})
            return self.to_df(tb[df.columns], schema=df.schema)

        pcols = ", ".join(
            self.encode_column_name(x) for x in partition_spec.partition_by
        )
        sql = (
            f"SELECT * FROM ("
            f"SELECT *, ROW_NUMBER() OVER (PARTITION BY {pcols} {sort_expr}) "
            f"AS __fugue_take_param FROM {tbn}"
            f") WHERE __fugue_take_param<={n}"
        )
        tb = self.query_to_table(sql, {tbn: idf})
        return self.to_df(tb[df.columns], schema=df.schema)

    def table_exists(self, table: str) -> bool:
        return table in self.backend.list_tables()

    def save_table(
        self,
        df: DataFrame,
        table: str,
        mode: str = "overwrite",
        partition_spec: Optional[PartitionSpec] = None,
        **kwargs: Any,
    ) -> None:
        if mode == "overwrite":
            self.backend.drop_table(table, force=True)
        if isinstance(df, IbisDataFrame):
            self.backend.create_table(table, df.native)
        else:
            self.backend.create_table(
                table, df.as_pandas(), schema=to_ibis_schema(df.schema)
            )

    def load_table(self, table: str, **kwargs: Any) -> DataFrame:
        return self.to_df(self.backend.table(table))


class IbisMapEngine(MapEngine):
    """IbisExecutionEngine's MapEngine, it is a wrapper of the map engine
    of :meth:`~.IbisExecutionEngine.non_ibis_engine`

    :param execution_engine: the execution engine this map engine will run on
    """

    @property
    def is_distributed(self) -> bool:
        return self._ibis_engine.non_ibis_engine.map_engine.is_distributed

    def __init__(self, execution_engine: ExecutionEngine) -> None:
        super().__init__(execution_engine)
        self._ibis_engine: IbisExecutionEngine = execution_engine  # type: ignore

    @property
    def execution_engine_constraint(self) -> Type[ExecutionEngine]:
        return IbisExecutionEngine

    def map_dataframe(
        self,
        df: DataFrame,
        map_func: Callable[[PartitionCursor, LocalDataFrame], LocalDataFrame],
        output_schema: Any,
        partition_spec: PartitionSpec,
        on_init: Optional[Callable[[int, DataFrame], Any]] = None,
        map_func_format_hint: Optional[str] = None,
    ) -> DataFrame:
        _df = self._ibis_engine._to_non_ibis_dataframe(df)
        return self._ibis_engine.non_ibis_engine.map_engine.map_dataframe(
            _df,
            map_func=map_func,
            output_schema=output_schema,
            partition_spec=partition_spec,
            on_init=on_init,
            map_func_format_hint=map_func_format_hint,
        )

    def map_bag(
        self,
        bag: Bag,
        map_func: Callable[[BagPartitionCursor, LocalBag], LocalBag],
        partition_spec: PartitionSpec,
        on_init: Optional[Callable[[int, Bag], Any]] = None,
    ) -> Bag:  # pragma: no cover
        return self._ibis_engine.non_ibis_engine.map_engine.map_bag(
            bag, map_func, partition_spec, on_init
        )


class IbisExecutionEngine(ExecutionEngine):
    """The base execution engine using Ibis.
    Please read |ExecutionEngineTutorial| to understand this important Fugue concept

    :param conf: |ParamsLikeObject|, read |FugueConfig| to learn Fugue specific options
    """

    def __init__(self, conf: Any):
        super().__init__(conf)
        self._non_ibis_engine = self.create_non_ibis_execution_engine()

    @abstractmethod
    def create_non_ibis_execution_engine(self) -> ExecutionEngine:
        """Create the execution engine that handles operations beyond SQL"""
        raise NotImplementedError  # pragma: no cover

    def is_non_ibis(self, ds: Any) -> bool:
        return not isinstance(ds, (IbisDataFrame, IbisTable))

    @property
    def non_ibis_engine(self) -> ExecutionEngine:
        return self._non_ibis_engine

    @property
    def ibis_sql_engine(self) -> IbisSQLEngine:
        return self.sql_engine  # type: ignore

    def create_default_map_engine(self) -> MapEngine:
        return IbisMapEngine(self)

    @property
    def log(self) -> logging.Logger:
        return self.non_ibis_engine.log

    def get_current_parallelism(self) -> int:
        return self.non_ibis_engine.get_current_parallelism()

    def to_df(self, df: Any, schema: Any = None) -> DataFrame:
        if self.is_non_ibis(df):
            return self._to_non_ibis_dataframe(df, schema)
        return self._to_ibis_dataframe(df, schema=schema)

    def repartition(self, df: DataFrame, partition_spec: PartitionSpec) -> DataFrame:
        if self.is_non_ibis(df):
            return self.non_ibis_engine.repartition(df, partition_spec=partition_spec)
        self.log.warning("%s doesn't respect repartition", self)
        return df

    def broadcast(self, df: DataFrame) -> DataFrame:
        if self.is_non_ibis(df):
            return self.non_ibis_engine.broadcast(df)
        return df

    def persist(
        self,
        df: DataFrame,
        lazy: bool = False,
        **kwargs: Any,
    ) -> DataFrame:
        if self.is_non_ibis(df):
            return self.non_ibis_engine.persist(df, lazy=lazy, **kwargs)
        return self.ibis_sql_engine.persist(df, lazy=lazy, **kwargs)

    def join(
        self,
        df1: DataFrame,
        df2: DataFrame,
        how: str,
        on: Optional[List[str]] = None,
    ) -> DataFrame:
        if self.is_non_ibis(df1) and self.is_non_ibis(df2):
            return self.non_ibis_engine.join(df1, df2, how=how, on=on)
        return self.ibis_sql_engine.join(df1, df2, how=how, on=on)

    def union(self, df1: DataFrame, df2: DataFrame, distinct: bool = True) -> DataFrame:
        if self.is_non_ibis(df1) and self.is_non_ibis(df2):
            return self.non_ibis_engine.union(df1, df2, distinct=distinct)
        return self.ibis_sql_engine.union(df1, df2, distinct=distinct)

    def subtract(
        self, df1: DataFrame, df2: DataFrame, distinct: bool = True
    ) -> DataFrame:
        if self.is_non_ibis(df1) and self.is_non_ibis(df2):
            return self.non_ibis_engine.subtract(df1, df2, distinct=distinct)
        return self.ibis_sql_engine.subtract(df1, df2, distinct=distinct)

    def intersect(
        self, df1: DataFrame, df2: DataFrame, distinct: bool = True
    ) -> DataFrame:
        if self.is_non_ibis(df1) and self.is_non_ibis(df2):
            return self.non_ibis_engine.intersect(df1, df2, distinct=distinct)
        return self.ibis_sql_engine.intersect(df1, df2, distinct=distinct)

    def distinct(self, df: DataFrame) -> DataFrame:
        if self.is_non_ibis(df):
            return self.non_ibis_engine.distinct(df)
        return self.ibis_sql_engine.distinct(df)

    def dropna(
        self,
        df: DataFrame,
        how: str = "any",
        thresh: int = None,
        subset: Optional[List[str]] = None,
    ) -> DataFrame:
        if self.is_non_ibis(df):
            return self.non_ibis_engine.dropna(
                df, how=how, thresh=thresh, subset=subset
            )
        return self.ibis_sql_engine.dropna(df, how=how, thresh=thresh, subset=subset)

    def fillna(self, df: DataFrame, value: Any, subset: List[str] = None) -> DataFrame:
        if self.is_non_ibis(df):
            return self.non_ibis_engine.fillna(df, value=value, subset=subset)
        return self.ibis_sql_engine.fillna(df, value=value, subset=subset)

    def sample(
        self,
        df: DataFrame,
        n: Optional[int] = None,
        frac: Optional[float] = None,
        replace: bool = False,
        seed: Optional[int] = None,
    ) -> DataFrame:
        if self.is_non_ibis(df):
            return self.non_ibis_engine.sample(
                df,
                n=n,
                frac=frac,
                replace=replace,
                seed=seed,
            )
        return self.ibis_sql_engine.sample(
            df,
            n=n,
            frac=frac,
            replace=replace,
            seed=seed,
        )

    def take(
        self,
        df: DataFrame,
        n: int,
        presort: str,
        na_position: str = "last",
        partition_spec: Optional[PartitionSpec] = None,
    ) -> DataFrame:
        if self.is_non_ibis(df):
            return self.non_ibis_engine.take(
                df,
                n=n,
                presort=presort,
                na_position=na_position,
                partition_spec=partition_spec,
            )
        return self.ibis_sql_engine.take(
            df,
            n=n,
            presort=presort,
            na_position=na_position,
            partition_spec=partition_spec,
        )

    def _to_ibis_dataframe(self, df: Any, schema: Any = None) -> DataFrame:
        """Create ``IbisDataFrame`` from the dataframe like input

        :param df: dataframe like object
        :param schema: dataframe schema, defaults to None
        :return: the IbisDataFrame
        """
        return self.sql_engine.to_df(df, schema=schema)

    def _to_non_ibis_dataframe(
        self, df: Any, schema: Any = None
    ) -> DataFrame:  # pragma: no cover
        """Create ``DataFrame`` for map operations
        from the dataframe like input

        :param df: dataframe like object
        :param schema: dataframe schema, defaults to None
        :return: the DataFrame
        """
        return self.non_ibis_engine.to_df(df, schema)



import importlib

from .contrib import FUGUE_CONTRIB


def load_namespace(namespace: str) -> None:
    if namespace in FUGUE_CONTRIB:
        path = FUGUE_CONTRIB[namespace]["module"]
        importlib.import_module(path)



# pylint: disable-all
try:
    import qpd_pandas  # noqa: F401

    HAS_QPD = True
except ImportError:  # pragma: no cover
    HAS_QPD = False

import datetime
import os
import pickle
from typing import Any, Callable, Dict, Iterable, Iterator, List, Optional
from unittest import TestCase
from uuid import uuid4
from triad.utils.io import write_text, join
import numpy as np
import pandas as pd
import pyarrow as pa
import pytest
from fsspec.implementations.local import LocalFileSystem
from pytest import raises
from triad import SerializableRLock

import fugue.api as fa
from fugue import (
    AnyDataFrame,
    ArrayDataFrame,
    CoTransformer,
    DataFrame,
    DataFrames,
    ExecutionEngine,
    FugueWorkflow,
    LocalDataFrame,
    OutputCoTransformer,
    Outputter,
    OutputTransformer,
    PandasDataFrame,
    PartitionSpec,
    Processor,
    QPDPandasEngine,
    Schema,
    Transformer,
    cotransformer,
    output_cotransformer,
    output_transformer,
    outputter,
    processor,
    register_creator,
    register_default_sql_engine,
    register_output_transformer,
    register_outputter,
    register_processor,
    register_transformer,
    transformer,
)
from fugue.column import col
from fugue.column import functions as ff
from fugue.column import lit
from fugue.dataframe.utils import _df_eq as df_eq
from fugue.exceptions import (
    FugueInterfacelessError,
    FugueWorkflowCompileError,
    FugueWorkflowCompileValidationError,
    FugueWorkflowError,
    FugueWorkflowRuntimeValidationError,
)

_LOCAL_FS = LocalFileSystem(auto_mkdir=True)


class BuiltInTests(object):
    """Workflow level general test suite. It is a more general end to end
    test suite than :class:`~fugue_test.execution_suite.ExecutionEngineTests`.
    Any new :class:`~fugue.execution.execution_engine.ExecutionEngine`
    should also pass this test suite.

    Whenever you add method to FugueWorkflow and WorkflowDataFrame, you should
    add correspondent tests here
    """

    class Tests(TestCase):
        @classmethod
        def setUpClass(cls):
            register_default_sql_engine(lambda engine: engine.sql_engine)
            cls._engine = cls.make_engine(cls)

        @property
        def engine(self) -> ExecutionEngine:
            return self._engine  # type: ignore

        @classmethod
        def tearDownClass(cls):
            cls._engine.stop()

        def make_engine(self) -> ExecutionEngine:  # pragma: no cover
            raise NotImplementedError

        def test_workflows(self):
            a = FugueWorkflow().df([[0]], "a:int")
            df_eq(a.compute(self.engine), [[0]], "a:int")

        def test_create_show(self):
            with FugueWorkflow() as dag:
                dag.df([[0]], "a:int").persist().partition(num=2).show()
                dag.df(dag.df([[0]], "a:int")).persist().broadcast().show(title="t")
            dag.run(self.engine)

        def test_create_df_equivalence(self):
            ndf = fa.as_fugue_engine_df(self.engine, pd.DataFrame([[0]], columns=["a"]))
            dag1 = FugueWorkflow()
            dag1.df(ndf).show()
            dag2 = FugueWorkflow()
            dag2.create(ndf).show()
            assert dag1.spec_uuid() == dag2.spec_uuid()

        def test_checkpoint(self):
            with raises(FugueWorkflowError):
                with FugueWorkflow() as dag:
                    dag.df([[0]], "a:int").checkpoint()
                dag.run(self.engine)

            self.engine.conf["fugue.workflow.checkpoint.path"] = os.path.join(
                self.tmpdir, "ck"
            )
            with FugueWorkflow() as dag:
                a = dag.df([[0]], "a:int").checkpoint()
                dag.df([[0]], "a:int").assert_eq(a)
            dag.run(self.engine)

        def test_deterministic_checkpoint(self):
            self.engine.conf["fugue.workflow.checkpoint.path"] = os.path.join(
                self.tmpdir, "ck"
            )
            temp_file = os.path.join(self.tmpdir, "t.parquet")

            def mock_create(dummy: int = 1) -> pd.DataFrame:
                return pd.DataFrame(np.random.rand(3, 2), columns=["a", "b"])

            # base case without checkpoint, two runs should generate different result
            with FugueWorkflow() as dag:
                a = dag.create(mock_create)
                a.save(temp_file)
            dag.run(self.engine)
            with FugueWorkflow() as dag:
                a = dag.create(mock_create)
                b = dag.load(temp_file)
                b.assert_not_eq(a)
            dag.run(self.engine)

            # strong checkpoint is not cross execution, so result should be different
            with FugueWorkflow() as dag:
                a = dag.create(mock_create).strong_checkpoint()
                a.save(temp_file)
            dag.run(self.engine)
            with FugueWorkflow() as dag:
                a = dag.create(mock_create).strong_checkpoint()
                b = dag.load(temp_file)
                b.assert_not_eq(a)
            dag.run(self.engine)

            # with deterministic checkpoint, two runs should generate the same result
            with FugueWorkflow() as dag:
                dag.create(
                    mock_create, params=dict(dummy=2)
                )  # this should not affect a because it's not a dependency
                a = dag.create(mock_create).deterministic_checkpoint()
                id1 = a.spec_uuid()
                a.save(temp_file)
            dag.run(self.engine)
            with FugueWorkflow() as dag:
                a = dag.create(mock_create).deterministic_checkpoint()
                b = dag.load(temp_file)
                b.assert_eq(a)
            dag.run(self.engine)
            with FugueWorkflow() as dag:
                # checkpoint specs itself doesn't change determinism
                # of the previous steps
                a = dag.create(mock_create).deterministic_checkpoint(
                    partition=PartitionSpec(num=2)
                )
                id2 = a.spec_uuid()
                b = dag.load(temp_file)
                b.assert_eq(a)
            dag.run(self.engine)
            with FugueWorkflow() as dag:
                # dependency change will affect determinism
                a = dag.create(
                    mock_create, params={"dummy": 2}
                ).deterministic_checkpoint()
                id3 = a.spec_uuid()
                b = dag.load(temp_file)
                b.assert_not_eq(a)
            dag.run(self.engine)

            assert (
                id1 == id2
            )  # different types of checkpoint doesn't change dag determinism
            assert id1 != id3

        def test_deterministic_checkpoint_complex_dag(self):
            self.engine.conf["fugue.workflow.checkpoint.path"] = os.path.join(
                self.tmpdir, "ck"
            )
            temp_file = os.path.join(self.tmpdir, "t.parquet")

            def mock_create(dummy: int = 1) -> pd.DataFrame:
                return pd.DataFrame(np.random.rand(3, 2), columns=["a", "b"])

            # base case without checkpoint, two runs should generate different result
            with FugueWorkflow() as dag:
                a = dag.create(mock_create).drop(["a"])
                b = dag.create(mock_create).drop(["a"])
                c = a.union(b)
                c.save(temp_file)
            dag.run(self.engine)
            with FugueWorkflow() as dag:
                a = dag.create(mock_create).drop(["a"])
                b = dag.create(mock_create).drop(["a"])
                c = a.union(b)
                d = dag.load(temp_file)
                d.assert_not_eq(c)
            dag.run(self.engine)

            # strong checkpoint is not cross execution, so result should be different
            with FugueWorkflow() as dag:
                a = dag.create(mock_create).drop(["a"])
                b = dag.create(mock_create).drop(["a"])
                c = a.union(b).strong_checkpoint()
                c.save(temp_file)
            dag.run(self.engine)
            with FugueWorkflow() as dag:
                a = dag.create(mock_create).drop(["a"])
                b = dag.create(mock_create).drop(["a"])
                c = a.union(b).strong_checkpoint()
                d = dag.load(temp_file)
                d.assert_not_eq(c)
            dag.run(self.engine)

            # with deterministic checkpoint, two runs should generate the same result
            with FugueWorkflow() as dag:
                dag.create(
                    mock_create, params=dict(dummy=2)
                )  # this should not affect a because it's not a dependency
                a = dag.create(mock_create).drop(["a"])
                b = dag.create(mock_create).drop(["a"])
                c = a.union(b)
                c.deterministic_checkpoint()
                c.save(temp_file)
            dag.run(self.engine)
            with FugueWorkflow() as dag:
                a = dag.create(mock_create).drop(["a"])
                b = dag.create(mock_create).drop(["a"])
                c = a.union(b).deterministic_checkpoint()
                d = dag.load(temp_file)
                d.assert_eq(c)
            dag.run(self.engine)
            with FugueWorkflow() as dag:
                # checkpoint specs itself doesn't change determinism
                # of the previous steps
                a = dag.create(mock_create).drop(["a"])
                b = dag.create(mock_create).drop(["a"])
                c = a.union(b).deterministic_checkpoint(partition=PartitionSpec(num=2))
                d = dag.load(temp_file)
                d.assert_eq(c)
            dag.run(self.engine)
            with FugueWorkflow() as dag:
                # dependency change will affect determinism
                a = dag.create(mock_create, params={"dummy": 2}).drop(["a"])
                b = dag.create(mock_create).drop(["a"])
                c = a.union(b).deterministic_checkpoint(partition=PartitionSpec(num=2))
                d = dag.load(temp_file)
                d.assert_not_eq(c)
            dag.run(self.engine)

        def test_yield_file(self):
            self.engine.conf["fugue.workflow.checkpoint.path"] = os.path.join(
                self.tmpdir, "ck"
            )

            with raises(FugueWorkflowCompileError):
                FugueWorkflow().df([[0]], "a:int").checkpoint().yield_file_as("x")

            with raises(ValueError):
                FugueWorkflow().df([[0]], "a:int").persist().yield_file_as("x")

            def run_test(deterministic):
                dag1 = FugueWorkflow()
                df = dag1.df([[0]], "a:int")
                if deterministic:
                    df = df.deterministic_checkpoint(storage_type="file")
                df.yield_file_as("x")
                id1 = dag1.spec_uuid()
                dag2 = FugueWorkflow()
                dag2.df([[0]], "a:int").assert_eq(dag2.df(dag1.yields["x"]))
                id2 = dag2.spec_uuid()
                dag1.run(self.engine)
                dag2.run(self.engine)
                return id1, id2

            id1, id2 = run_test(False)
            id3, id4 = run_test(False)
            assert id1 == id3
            assert id2 != id4  # non deterministic yield (direct yield)

            id1, id2 = run_test(True)
            id3, id4 = run_test(True)
            assert id1 == id3
            assert id2 == id4  # deterministic yield (yield deterministic checkpoint)

        def test_yield_table(self):
            with raises(FugueWorkflowCompileError):
                FugueWorkflow().df([[0]], "a:int").checkpoint().yield_table_as("x")

            with raises(ValueError):
                FugueWorkflow().df([[0]], "a:int").persist().yield_table_as("x")

            def run_test(deterministic):
                dag1 = FugueWorkflow()
                df = dag1.df([[0]], "a:int")
                if deterministic:
                    df = df.deterministic_checkpoint(storage_type="table")
                df.yield_table_as("x")
                id1 = dag1.spec_uuid()
                dag2 = FugueWorkflow()
                dag2.df([[0]], "a:int").assert_eq(dag2.df(dag1.yields["x"]))
                id2 = dag2.spec_uuid()
                dag1.run(self.engine)
                dag2.run(self.engine)
                return id1, id2

            id1, id2 = run_test(False)
            id3, id4 = run_test(False)
            assert id1 == id3
            assert id2 != id4  # non deterministic yield (direct yield)

            id1, id2 = run_test(True)
            id3, id4 = run_test(True)
            assert id1 == id3
            assert id2 == id4  # deterministic yield (yield deterministic checkpoint)

        def test_yield_dataframe(self):
            dag = FugueWorkflow()
            df1 = dag.df([[1]], "a:int")
            df2 = dag.df([[1]], "a:int")
            df1.union(df2, distinct=False).yield_dataframe_as("x")
            df1.union(df2, distinct=False).yield_dataframe_as("y", as_local=True)
            result = dag.run()["x"]
            assert [[1], [1]] == result.as_array()
            result = dag.run()["y"]
            assert [[1], [1]] == result.as_array()
            assert result.is_local

        def test_create_process_output(self):
            with FugueWorkflow() as dag:
                a = dag.create(mock_creator, params=dict(p=2))
                a.assert_eq(ArrayDataFrame([[2]], "a:int"))
                b = dag.process(a, a, using=mock_processor)
                b.assert_eq(ArrayDataFrame([[2]], "a:int"))
                b = dag.process(dict(df1=a, df2=a), using=mock_processor)
                b.assert_eq(ArrayDataFrame([[2]], "a:int"))
                dag.output(a, b, using=mock_outputter)
                b2 = dag.process(a, a, a, using=mock_processor2)
                b2.assert_eq(ArrayDataFrame([[3]], "a:int"))
                b2 = dag.process(a, a, a, using=MockProcessor3)
                b2.assert_eq(ArrayDataFrame([[3]], "a:int"))
                a.process(mock_processor2).assert_eq(ArrayDataFrame([[1]], "a:int"))
                a.output(mock_outputter2)
                dag.output(dict(df=a), using=mock_outputter2)
                a.partition(num=3).output(MockOutputter3)
                dag.output(dict(aa=a, bb=b), using=MockOutputter4)

                a = dag.create(mock_creator2, params=dict(p=2))
                b = dag.create(mock_creator2, params=dict(p=2))
                c = dag.process(a, b, using=mock_processor4)
                c.assert_eq(ArrayDataFrame([[2]], "a:int"))
                dag.output(a, b, using=mock_outputter4)
            dag.run(self.engine)

        def test_zip(self):
            with FugueWorkflow() as dag:
                a = dag.df([[1, 2], [2, 3], [2, 5]], "a:int,b:int")
                b = dag.df([[1, 3]], "a:int,c:int")
                c1 = a.zip(b)
                c2 = dag.zip(a, b)
                c1.assert_eq(c2)

                a = dag.df([[1, 2], [2, 3], [2, 5]], "a:int,b:int")
                b = dag.df([[1, 3]], "a:int,c:int")
                c1 = a.zip(b, how="left_outer", partition=dict(presort="b DESC, c ASC"))
                c2 = dag.zip(
                    a, b, how="left_outer", partition=dict(presort="b DESC, c ASC")
                )
                c1.assert_eq(c2)

                a = dag.df([[1, 2, 0], [1, 3, 1]], "a:int,b:int,c:int")
                b = dag.df([[1, 2, 1], [1, 3, 2]], "a:int,b:int,d:int")
                c = dag.df([[1, 4]], "a:int,e:int")
                e = dag.df([[1]], "a:int")
                dag.zip(a, b, c)[["a"]].distinct().assert_eq(e)
            dag.run(self.engine)

        def test_transform(self):
            with FugueWorkflow() as dag:
                a = dag.df([[1, 2], [3, 4]], "a:double,b:int")
                c = a.transform(mock_tf0)
                dag.df([[1, 2, 1], [3, 4, 1]], "a:double,b:int,p:int").assert_eq(c)

                c = a.transform(mock_tf0, params=dict(col="x"))
                dag.df([[1, 2, 1], [3, 4, 1]], "a:double,b:int,x:int").assert_eq(c)

                a = dag.df([[1, 2], [None, 1], [3, 4], [None, 4]], "a:double,b:int")
                c = a.transform(mock_tf0, params=dict(p="10"))
                dag.df(
                    [[1, 2, 10], [None, 1, 10], [3, 4, 10], [None, 4, 10]],
                    "a:double,b:int,p:int",
                ).assert_eq(c)
            dag.run(self.engine)

        def test_local_instance_as_extension(self):
            class _Mock(object):
                # schema: *
                def t1(self, df: pd.DataFrame) -> pd.DataFrame:
                    return df

                def t2(self, df: pd.DataFrame) -> pd.DataFrame:
                    return df

                def test(self):
                    with FugueWorkflow() as dag_:
                        a = dag_.df([[0], [1]], "a:int")
                        b = a.transform(self.t1)
                        b.assert_eq(a)
                    dag_.run()

            m = _Mock()
            m.test()
            with FugueWorkflow() as dag:
                a = dag.df([[0], [1]], "a:int")
                b = a.transform(m.t1).transform(m.t2, schema="*")
                b.assert_eq(a)
            dag.run(self.engine)

        def test_transform_iterable_dfs(self):
            # this test is important for using mapInPandas in spark

            # schema: *,c:int
            def mt_pandas(
                dfs: Iterable[pd.DataFrame], empty: bool = False
            ) -> Iterator[pd.DataFrame]:
                for df in dfs:
                    if not empty:
                        df = df.assign(c=2)
                        df = df[reversed(list(df.columns))]
                        yield df

            with FugueWorkflow() as dag:
                a = dag.df([[1, 2], [3, 4]], "a:int,b:int")
                b = a.transform(mt_pandas)
                dag.df([[1, 2, 2], [3, 4, 2]], "a:int,b:int,c:int").assert_eq(b)
            dag.run(self.engine)

            # when iterable returns nothing
            with FugueWorkflow() as dag:
                a = dag.df([[1, 2], [3, 4]], "a:int,b:int")
                # without partitioning
                b = a.transform(mt_pandas, params=dict(empty=True))
                dag.df([], "a:int,b:int,c:int").assert_eq(b)
                # with partitioning
                b = a.partition_by("a").transform(mt_pandas, params=dict(empty=True))
                dag.df([], "a:int,b:int,c:int").assert_eq(b)
            dag.run(self.engine)

            # schema: *
            def mt_arrow(
                dfs: Iterable[pa.Table], empty: bool = False
            ) -> Iterator[pa.Table]:
                for df in dfs:
                    if not empty:
                        df = df.select(reversed(df.schema.names))
                        yield df

            # schema: a:long
            def mt_arrow_2(dfs: Iterable[pa.Table]) -> Iterator[pa.Table]:
                for df in dfs:
                    yield df.drop(["b"])

            with FugueWorkflow() as dag:
                a = dag.df([[1, 2], [3, 4]], "a:int,b:int")
                b = a.transform(mt_arrow)
                dag.df([[1, 2], [3, 4]], "a:int,b:int").assert_eq(b)
                b = a.transform(mt_arrow_2)
                dag.df([[1], [3]], "a:long").assert_eq(b)
            dag.run(self.engine)

            # when iterable returns nothing
            with FugueWorkflow() as dag:
                a = dag.df([[1, 2], [3, 4]], "a:int,b:int")
                # without partitioning
                b = a.transform(mt_arrow, params=dict(empty=True))
                dag.df([], "a:int,b:int").assert_eq(b)
                # with partitioning
                b = a.partition_by("a").transform(mt_arrow, params=dict(empty=True))
                dag.df([], "a:int,b:int").assert_eq(b)
            dag.run(self.engine)

        def test_transform_binary(self):
            with FugueWorkflow() as dag:
                a = dag.df([[1, pickle.dumps([0, "a"])]], "a:int,b:bytes")
                c = a.transform(mock_tf3).persist()
                b = dag.df([[1, pickle.dumps([1, "ax"])]], "a:int,b:bytes")
                b.assert_eq(c, check_order=True)
            dag.run(self.engine)

        def test_transform_by(self):
            with FugueWorkflow() as dag:
                a = dag.df([[1, 2], [None, 1], [3, 4], [None, 4]], "a:double,b:int")
                c = a.transform(MockTransform1, pre_partition={"by": ["a"]})
                dag.df(
                    [[None, 1, 2, 1], [None, 4, 2, 1], [1, 2, 1, 1], [3, 4, 1, 1]],
                    "a:double,b:int,ct:int,p:int",
                ).assert_eq(c)

                c = a.transform(
                    mock_tf1, pre_partition={"by": ["a"], "presort": "b DESC"}
                )
                dag.df(
                    [[None, 4, 2, 1], [None, 1, 2, 1], [1, 2, 1, 1], [3, 4, 1, 1]],
                    "a:double,b:int,ct:int,p:int",
                ).assert_eq(c)

                c = a.transform(
                    mock_tf2_except,
                    schema="*",
                    pre_partition={"by": ["a"], "presort": "b DESC"},
                    ignore_errors=[NotImplementedError],
                )
                dag.df([[1, 2], [3, 4]], "a:double,b:int").assert_eq(c)

                c = a.partition(by="a", presort="b DESC").transform(
                    mock_tf2_except, schema="*", ignore_errors=[NotImplementedError]
                )
                dag.df([[1, 2], [3, 4]], "a:double,b:int").assert_eq(c)
            dag.run(self.engine)

        def test_cotransform(self):
            with FugueWorkflow() as dag:
                a = dag.df([[1, 2], [1, 3], [2, 1]], "a:int,b:int")
                b = dag.df([[1, 2], [3, 4]], "a:int,c:int")
                c = dag.transform(a.zip(b), using=MockCoTransform1)
                e = dag.df([[1, 2, 1, 1]], "a:int,ct1:int,ct2:int,p:int")
                e.assert_eq(c)
                c = dag.transform(a.zip(b), using=MockCoTransform1, params=dict(p=10))
                e = dag.df([[1, 2, 1, 10]], "a:int,ct1:int,ct2:int,p:int")
                e.assert_eq(c)
                # with deco
                c = dag.transform(a.zip(b), using=mock_co_tf1, params=dict(p=10))
                e = dag.df([[1, 2, 1, 10]], "a:int,ct1:int,ct2:int,p:int")
                e.assert_eq(c)

                a.zip(b).transform(mock_co_tf1, params=dict(p=10)).assert_eq(e)

                c = dag.transform(
                    a.zip(b), using=mock_co_tf1, params=dict(p=10, col="x")
                )
                e = dag.df([[1, 2, 1, 10]], "a:int,ct1:int,ct2:int,x:int")
                e.assert_eq(c)

                # interfaceless
                c = dag.transform(
                    a.zip(b),
                    using=mock_co_tf2,
                    schema="a:int,ct1:int,ct2:int,p:int",
                    params=dict(p=10),
                )
                e = dag.df([[1, 2, 1, 10]], "a:int,ct1:int,ct2:int,p:int")
                e.assert_eq(c)
                # single df
                c = dag.transform(a.zip(), using=mock_co_tf3)
                e = dag.df([[1, 3, 1]], "a:int,ct1:int,p:int")
                e.assert_eq(c)
                c = dag.transform(a.zip(partition=dict(by=["a"])), using=mock_co_tf3)
                e = dag.df([[1, 2, 1], [2, 1, 1]], "a:int,ct1:int,p:int")
                e.assert_eq(c)
                c = a.partition_by("a").zip().transform(mock_co_tf3)
                e = dag.df([[1, 2, 1], [2, 1, 1]], "a:int,ct1:int,p:int")
                e.assert_eq(c)
                # ignore errors
                c = (
                    a.partition(by=["a"])
                    .zip()
                    .transform(mock_co_tf4_ex, ignore_errors=[NotImplementedError])
                )
                e = dag.df([[1, 2, 1]], "a:int,ct1:int,p:int")
                e.assert_eq(c)
            dag.run(self.engine)

        def test_cotransform_with_key(self):
            with FugueWorkflow() as dag:
                a = dag.df([[1, 2], [1, 3], [2, 1]], "a:int,b:int")
                b = dag.df([[1, 2], [3, 4]], "a:int,c:int")
                dag.zip(dict(x=a, y=b)).show()
                c = dag.zip(dict(x=a, y=b)).transform(
                    MockCoTransform1, params=dict(named=True)
                )
                e = dag.df([[1, 2, 1, 1]], "a:int,ct1:int,ct2:int,p:int")
                e.assert_eq(c)

                c = dag.zip(dict(df1=a, df2=b)).transform(
                    mock_co_tf1, params=dict(p=10)
                )
                e = dag.df([[1, 2, 1, 10]], "a:int,ct1:int,ct2:int,p:int")
                e.assert_eq(c)

                c = dag.zip(dict(df2=a, df1=b)).transform(
                    mock_co_tf1, params=dict(p=10)
                )
                e = dag.df([[1, 1, 2, 10]], "a:int,ct1:int,ct2:int,p:int")
                e.assert_eq(c)

                c = dag.transform(
                    dag.zip(dict(x=a, y=b)),
                    using=mock_co_tf2,
                    schema="a:int,ct1:int,ct2:int,p:int",
                    params=dict(p=10),
                )
                e = dag.df([[1, 2, 1, 10]], "a:int,ct1:int,ct2:int,p:int")
                e.assert_eq(c)

                c = dag.zip(dict(df1=a)).transform(mock_co_tf3)
                e = dag.df([[1, 3, 1]], "a:int,ct1:int,p:int")
                e.assert_eq(c)
            dag.run(self.engine)

        def test_out_transform(self):  # noqa: C901
            tmpdir = str(self.tmpdir)

            def incr():
                write_text(join(tmpdir, str(uuid4()) + ".txt"), "")
                return len(_LOCAL_FS.glob(join(tmpdir, "*.txt")))

            def t1(df: Iterable[Dict[str, Any]]) -> Iterable[Dict[str, Any]]:
                for row in df:
                    incr()
                    yield row

            # partitionby_has: b
            def t2(df: pd.DataFrame) -> None:
                incr()

            # schema: *
            # partitionby_has: b
            def t3(df: pd.DataFrame) -> pd.DataFrame:
                incr()
                return df

            @transformer("*", partitionby_has=["b"])
            def t4(df: Iterable[Dict[str, Any]]) -> Iterable[Dict[str, Any]]:
                for row in df:
                    incr()
                    yield row

            @output_transformer(partitionby_has=["b"])
            def t5(df: Iterable[Dict[str, Any]]) -> Iterable[Dict[str, Any]]:
                for row in df:
                    incr()
                    yield row

            class T6(Transformer):
                def get_output_schema(self, df):
                    return df.schema

                def transform(self, df):
                    incr()
                    return df

            class T7(OutputTransformer):
                @property
                def validation_rules(self):
                    return {"partitionby_has": "b"}

                def process(self, df):
                    incr()

            def t8(df: pd.DataFrame) -> None:
                incr()
                raise NotImplementedError

            def t9(df: pd.DataFrame) -> Iterable[pd.DataFrame]:
                incr()
                yield df

            def t10(df: pd.DataFrame) -> Iterable[pa.Table]:
                incr()
                yield pa.Table.from_pandas(df)

            with FugueWorkflow() as dag:
                a = dag.df([[1, 2], [3, 4]], "a:double,b:int")
                a.out_transform(t1)  # +2
                a.partition_by("b").out_transform(t2)  # +1 or +2
                a.partition_by("b").out_transform(t3)  # +1 or +2
                a.partition_by("b").out_transform(t4)  # +2
                a.partition_by("b").out_transform(t5)  # +2
                a.out_transform(T6)  # +1
                a.partition_by("b").out_transform(T7)  # +1
                a.out_transform(t8, ignore_errors=[NotImplementedError])  # +1
                a.out_transform(t9)  # +1
                a.out_transform(t10)  # +1
                raises(FugueWorkflowCompileValidationError, lambda: a.out_transform(t2))
                raises(FugueWorkflowCompileValidationError, lambda: a.out_transform(t3))
                raises(FugueWorkflowCompileValidationError, lambda: a.out_transform(t4))
                raises(FugueWorkflowCompileValidationError, lambda: a.out_transform(t5))
                raises(FugueWorkflowCompileValidationError, lambda: a.out_transform(T7))
            dag.run(self.engine)

            assert 13 <= incr()

        def test_out_cotransform(self):  # noqa: C901
            tmpdir = str(self.tmpdir)

            def incr():
                write_text(join(tmpdir, str(uuid4()) + ".txt"), "")
                return len(_LOCAL_FS.glob(join(tmpdir, "*.txt")))

            def t1(
                df: Iterable[Dict[str, Any]], df2: pd.DataFrame
            ) -> Iterable[Dict[str, Any]]:
                for row in df:
                    incr()
                    yield row

            def t2(dfs: DataFrames) -> None:
                incr()

            def t3(df: pd.DataFrame, df2: pd.DataFrame) -> pd.DataFrame:
                incr()
                return df

            @cotransformer("a:double,b:int")
            def t4(
                df: Iterable[Dict[str, Any]], df2: pd.DataFrame
            ) -> Iterable[Dict[str, Any]]:
                for row in df:
                    incr()
                    yield row

            @output_cotransformer()
            def t5(
                df: Iterable[Dict[str, Any]], df2: pd.DataFrame
            ) -> Iterable[Dict[str, Any]]:
                for row in df:
                    incr()
                    yield row

            class T6(CoTransformer):
                def get_output_schema(self, dfs):
                    return dfs[0].schema

                def transform(self, dfs):
                    incr()
                    return dfs[0]

            class T7(OutputCoTransformer):
                def process(self, dfs):
                    incr()

            def t8(df: pd.DataFrame, df2: pd.DataFrame) -> pd.DataFrame:
                incr()
                raise NotImplementedError

            def t9(df1: pd.DataFrame, df2: pd.DataFrame) -> Iterable[pd.DataFrame]:
                incr()
                for df in [df1, df2]:
                    yield df

            with FugueWorkflow() as dag:
                a0 = dag.df([[1, 2], [3, 4]], "a:double,b:int")
                a1 = dag.df([[1, 2], [3, 4]], "aa:double,b:int")
                a = dag.zip(a0, a1)
                a.out_transform(t1)  # +2
                a.out_transform(t2)  # +1 or +2
                a.out_transform(t3)  # +1 or +2
                a.out_transform(t4)  # +2
                b = dag.zip(dict(df=a0, df2=a1))
                b.out_transform(t5)  # +2
                a.out_transform(T6)  # +1
                a.out_transform(T7)  # +1
                a.out_transform(t8, ignore_errors=[NotImplementedError])  # +1
                a.out_transform(t9)  # +1
            dag.run(self.engine)

            assert 12 <= incr()

        def test_join(self):
            with FugueWorkflow() as dag:
                a = dag.df([[1, 10], [2, 20], [3, 30]], "a:int,b:int")
                dag.join(a, how="inner").assert_eq(a)

                b = dag.df([[2, 200], [3, 300]], "a:int,c:int")
                c = dag.df([[2, 2000]], "a:int,d:int")
                d = a.join(b, c, how="inner")  # infer join on
                dag.df([[2, 20, 200, 2000]], "a:int,b:int,c:int,d:int").assert_eq(d)
                d = a.inner_join(b, c)
                dag.df([[2, 20, 200, 2000]], "a:int,b:int,c:int,d:int").assert_eq(d)
                d = a.semi_join(b, c)
                dag.df([[2, 20]], "a:int,b:int").assert_eq(d)
                d = a.left_semi_join(b, c)
                dag.df([[2, 20]], "a:int,b:int").assert_eq(d)
                d = a.anti_join(b, c)
                dag.df([[1, 10]], "a:int,b:int").assert_eq(d)
                d = a.left_anti_join(b, c)
                dag.df([[1, 10]], "a:int,b:int").assert_eq(d)

                # TODO: change these to str type to only test outer features?
                a = dag.df([[1, 10], [2, 20], [3, 30]], "a:int,b:int")
                b = dag.df([[2, 200], [3, 300]], "a:int,c:int")
                c = dag.df([[2, 2000], [4, 4000]], "a:int,d:int")
                d = a.left_outer_join(b, c)
                d.show()
                dag.df(
                    [[1, 10, None, None], [2, 20, 200, 2000], [3, 30, 300, None]],
                    "a:int,b:int,c:int,d:int",
                ).assert_eq(d)
                d = a.right_outer_join(b, c)
                dag.df(
                    [[2, 20, 200, 2000], [4, None, None, 4000]],
                    "a:int,b:int,c:int,d:int",
                ).assert_eq(d)
                d = a.full_outer_join(b, c)
                dag.df(
                    [
                        [1, 10, None, None],
                        [2, 20, 200, 2000],
                        [3, 30, 300, None],
                        [4, None, None, 4000],
                    ],
                    "a:int,b:int,c:int,d:int",
                ).assert_eq(d)

                a = dag.df([[1, 10], [2, 20]], "a:int,b:int")
                b = dag.df([[2], [3]], "c:int")
                c = dag.df([[4]], "d:int")
                d = a.cross_join(b, c)
                dag.df(
                    [[1, 10, 2, 4], [1, 10, 3, 4], [2, 20, 2, 4], [2, 20, 3, 4]],
                    "a:int,b:int,c:int,d:int",
                ).assert_eq(d)
            dag.run(self.engine)

        @pytest.mark.skipif(not HAS_QPD, reason="qpd not working")
        def test_df_select(self):
            with FugueWorkflow() as dag:
                # wildcard
                a = dag.df([[1, 10], [2, 20], [3, 30]], "x:int,y:int")
                a.select("*").assert_eq(a)

                # column transformation
                b = dag.df(
                    [[1, 10, 11, "x"], [2, 20, 22, "x"], [3, 30, 33, "x"]],
                    "x:int,y:int,c:int,d:str",
                )
                a.select(
                    "*",
                    (col("x") + col("y")).cast("int32").alias("c"),
                    lit("x", "d"),
                ).assert_eq(b)

                # distinct
                a = dag.df([[1, 10], [2, 20], [1, 10]], "x:int,y:int")
                b = dag.df([[1, 10], [2, 20]], "x:int,y:int")
                a.select("*", distinct=True).assert_eq(b)

                # aggregation plus infer alias
                a = dag.df([[1, 10], [1, 20], [3, 30]], "x:int,y:int")
                b = dag.df([[1, 30], [3, 30]], "x:int,y:int")
                a.select("x", ff.sum(col("y")).cast("int32")).assert_eq(b)

                # all together
                a = dag.df([[1, 10], [1, 20], [3, 35], [3, 40]], "x:int,y:int")
                b = dag.df([[3, 35]], "x:int,z:int")
                a.select(
                    "x",
                    ff.sum(col("y")).alias("z").cast("int32"),
                    where=col("y") < 40,
                    having=ff.sum(col("y")) > 30,
                ).assert_eq(b)

                b = dag.df([[65]], "z:long")
                a.select(
                    ff.sum(col("y")).alias("z").cast(int), where=col("y") < 40
                ).show()

                raises(ValueError, lambda: a.select("*", "x"))
            dag.run(self.engine)

            with FugueWorkflow() as dag:
                a = dag.df([[0]], "a:long")
                b = dag.df([[0]], "a:long")
                dag.select("select * from", a).assert_eq(b)
            dag.run(self.engine, {"fugue.sql.compile.ignore_case": True})

        @pytest.mark.skipif(not HAS_QPD, reason="qpd not working")
        def test_df_filter(self):
            with FugueWorkflow() as dag:
                a = dag.df([[1, 10], [2, 20], [3, 30]], "x:int,y:int")
                b = dag.df([[2, 20]], "x:int,y:int")
                a.filter((col("y") > 15) & (col("y") < 25)).assert_eq(b)
            dag.run(self.engine)

        @pytest.mark.skipif(not HAS_QPD, reason="qpd not working")
        def test_df_assign(self):
            with FugueWorkflow() as dag:
                a = dag.df([[1, 10], [2, 20], [3, 30]], "x:int,y:int")
                b = dag.df([[1, "x"], [2, "x"], [3, "x"]], "x:int,y:str")
                a.assign(y="x").assert_eq(b)

                a = dag.df([[1, 10], [2, 20], [3, 30]], "x:int,y:int")
                b = dag.df(
                    [[1, "x", 11], [2, "x", 21], [3, "x", 31]], "x:int,y:str,z:double"
                )
                a.assign(lit("x").alias("y"), z=(col("y") + 1).cast(float)).assert_eq(b)
            dag.run(self.engine)

        @pytest.mark.skipif(not HAS_QPD, reason="qpd not working")
        def test_aggregate(self):
            with FugueWorkflow() as dag:
                a = dag.df([[1, 10], [1, 200], [3, 30]], "x:int,y:int")
                b = dag.df([[1, 200], [3, 30]], "x:int,y:int")
                c = dag.df([[-200, 200, 70]], "y:int,zz:int,ww:int")
                a.partition_by("x").aggregate(ff.max(col("y"))).assert_eq(b)
                a.aggregate(
                    ff.min(-col("y")),
                    zz=ff.max(col("y")),
                    ww=((ff.min(col("y")) + ff.max(col("y"))) / 3).cast("int32"),
                ).assert_eq(c)
            dag.run(self.engine)

        @pytest.mark.skipif(not HAS_QPD, reason="qpd not working")
        def test_select(self):
            class MockEngine(QPDPandasEngine):
                def __init__(self, execution_engine, p: int = 0):
                    super().__init__(execution_engine)
                    self.p = p

                def select(self, dfs, statement):
                    assert 2 == self.p  # assert set p value is working
                    return super().select(dfs, statement)

            with FugueWorkflow() as dag:
                a = dag.df([[1, 10], [2, 20], [3, 30]], "x:long,y:long")
                b = dag.df([[2, 20, 40], [3, 30, 90]], "x:long,y:long,z:long")
                dag.select("* FROM", a).assert_eq(a)
                dag.select(a, ".* FROM", a).assert_eq(a)
                dag.select("SELECT *,x*y AS z FROM", a, "WHERE x>=2").assert_eq(b)

                c = dag.df([[2, 20, 40], [3, 30, 90]], "x:long,y:long,zb:long")
                dag.select(
                    "  SELECT t1.*,z AS zb FROM ",
                    a,
                    "AS t1 INNER JOIN",
                    b,
                    "AS t2 ON t1.x=t2.x  ",
                ).assert_eq(c)

                # no select
                dag.select(
                    "t1.*,z AS zb FROM ", a, "AS t1 INNER JOIN", b, "AS t2 ON t1.x=t2.x"
                ).assert_eq(c)

                # specify sql engine
                dag.select(
                    "SELECT t1.*,z AS zb FROM ",
                    a,
                    "AS t1 INNER JOIN",
                    b,
                    "AS t2 ON t1.x=t2.x",
                    sql_engine="qpdpandas",
                ).assert_eq(c)

                # specify sql engine and params
                dag.select(
                    "SELECT t1.*,z AS zb FROM ",
                    a,
                    "AS t1 INNER JOIN",
                    b,
                    "AS t2 ON t1.x=t2.x",
                    sql_engine=MockEngine,
                    sql_engine_params={"p": 2},
                ).assert_eq(c)

                # no input
                dag.select("'test' AS a").assert_eq(dag.df([["test"]], "a:str"))

                # make sure transform -> select works
                b = a.transform(mock_tf1)
                a = a.transform(mock_tf1)
                aa = dag.select("* FROM", a)
                dag.select("* FROM", b).assert_eq(aa)
            dag.run(self.engine)

        def test_union(self):
            with FugueWorkflow() as dag:
                a = dag.df([[1, 10], [2, None], [2, None]], "x:long,y:double")
                b = dag.df([[2, None], [2, 20]], "x:long,y:double")
                c = dag.df([[1, 10], [2, 20]], "x:long,y:double")
                a.union().assert_eq(a)
                a.union(b, c).assert_eq(
                    ArrayDataFrame(
                        [
                            [1, 10],
                            [2, None],
                            [2, 20],
                        ],
                        "x:long,y:double",
                    )
                )
                a.union(b, c, distinct=False).assert_eq(
                    ArrayDataFrame(
                        [
                            [1, 10],
                            [2, None],
                            [2, None],
                            [2, None],
                            [2, 20],
                            [1, 10],
                            [2, 20],
                        ],
                        "x:long,y:double",
                    )
                )
            dag.run(self.engine)

        def test_intersect(self):
            with FugueWorkflow() as dag:
                a = dag.df([[1, 10], [2, None], [2, None]], "x:long,y:double")
                b = dag.df([[2, None], [2, 20]], "x:long,y:double")
                c = dag.df([[1, 10], [2, 20]], "x:long,y:double")
                # d = dag.df([[1, 10], [2, 20], [2, None]], "x:long,y:double")
                a.intersect(b).assert_eq(
                    ArrayDataFrame(
                        [[2, None]],
                        "x:long,y:double",
                    )
                )
                a.intersect(b, c).assert_eq(
                    ArrayDataFrame(
                        [],
                        "x:long,y:double",
                    )
                )
                # TODO: INTERSECT ALL is not implemented (QPD issue)
                # a.intersect(b, distinct=False).assert_eq(
                #     ArrayDataFrame(
                #         [[2, None], [2, None]],
                #         "x:long,y:double",
                #     )
                # )
                # a.intersect(b, d, distinct=False).assert_eq(
                #     ArrayDataFrame(
                #         [[2, None], [2, None]],
                #         "x:long,y:double",
                #     )
                # )
            dag.run(self.engine)

        def test_subtract(self):
            with FugueWorkflow() as dag:
                a = dag.df([[1, 10], [2, None], [2, None]], "x:long,y:double")
                b = dag.df([[2, None], [2, 20]], "x:long,y:double")
                c = dag.df([[1, 10], [2, 20]], "x:long,y:double")
                a.subtract(b).assert_eq(
                    ArrayDataFrame(
                        [[1, 10]],
                        "x:long,y:double",
                    )
                )
                a.subtract(c).assert_eq(
                    ArrayDataFrame(
                        [[2, None]],
                        "x:long,y:double",
                    )
                )
                # # TODO: EXCEPT ALL is not implemented (QPD issue)
                # a.subtract(c, distinct=False).assert_eq(
                #     ArrayDataFrame(
                #         [[2, None], [2, None]],
                #         "x:long,y:double",
                #     )
                # )
                a.subtract(b, c).assert_eq(
                    ArrayDataFrame(
                        [],
                        "x:long,y:double",
                    )
                )
            dag.run(self.engine)

        def test_distinct(self):
            with FugueWorkflow() as dag:
                a = dag.df([[1, 10], [2, None], [2, None]], "x:long,y:double")
                a.distinct().assert_eq(
                    ArrayDataFrame(
                        [[1, 10], [2, None]],
                        "x:long,y:double",
                    )
                )
            dag.run(self.engine)

        def test_dropna(self):
            with FugueWorkflow() as dag:
                a = dag.df(
                    [[1, 10, 10], [None, 2, None], [2, None, 4]],
                    "x:double,y:double,z:double",
                )
                a.dropna().assert_eq(
                    ArrayDataFrame(
                        [[1, 10, 10]],
                        "x:double,y:double,z:double",
                    )
                )
                a.dropna(how="all").assert_eq(a)
                a.dropna(thresh=2).assert_eq(
                    ArrayDataFrame(
                        [[1, 10, 10], [2, None, 4]],
                        "x:double,y:double,z:double",
                    )
                )
                a.dropna(how="any", subset=["x", "z"]).assert_eq(
                    ArrayDataFrame(
                        [[1, 10, 10], [2, None, 4]],
                        "x:double,y:double,z:double",
                    )
                )
                a.dropna(thresh=1, subset=["y", "z"]).assert_eq(a)
            dag.run(self.engine)

        def test_fillna(self):
            with FugueWorkflow() as dag:
                a = dag.df(
                    [[1, 10, 10], [None, 2, None], [2, None, 4]],
                    "x:double,y:double,z:double",
                )
                a.fillna(-99).assert_eq(
                    ArrayDataFrame(
                        [[1, 10, 10], [-99, 2, -99], [2, -99, 4]],
                        "x:double,y:double,z:double",
                    )
                )
                a.fillna(-99, subset=["y"]).assert_eq(
                    ArrayDataFrame(
                        [[1, 10, 10], [None, 2, None], [2, -99, 4]],
                        "x:double,y:double,z:double",
                    )
                )
                # Subset is ignored if value is a dict
                a.fillna({"y": 0, "z": -99}, subset=["y"]).assert_eq(
                    ArrayDataFrame(
                        [[1, 10, 10], [None, 2, -99], [2, 0, 4]],
                        "x:double,y:double,z:double",
                    )
                )
            dag.run(self.engine)
            with raises(FugueWorkflowError):
                with FugueWorkflow() as dag:
                    dag.df([[None, 1]], "a:int,b:int").fillna({"a": None, "b": 1})
                dag.run(self.engine)

            with raises(FugueWorkflowError):
                with FugueWorkflow() as dag:
                    dag.df([[None, 1]], "a:int,b:int").fillna(None)
                dag.run(self.engine)

        def test_sample(self):
            with FugueWorkflow() as dag:
                a = dag.df(
                    [[1, 10, 10], [None, 2, None], [2, None, 4]],
                    "x:double,y:double,z:double",
                )
                a.sample(frac=0.5, replace=False, seed=0).show()
            dag.run(self.engine)

            with FugueWorkflow() as dag:
                with raises(ValueError):
                    dag.df([[None, 1]], "a:int,b:int").sample(n=1, frac=0.2)
            dag.run(self.engine)

        def test_take(self):
            # Test for presort parsing
            with FugueWorkflow() as dag:
                df = dag.df(
                    pd.DataFrame({"aaa": [1, 1, 2, 2], "bbb": ["a", "b", "c", "d"]})
                )
                df = df.partition(by=["aaa"], presort="bbb").take(1)
                df.show()
            dag.run(self.engine)
            # Partition but no presort. Output not deterministic
            with FugueWorkflow() as dag:
                df = dag.df(
                    pd.DataFrame({"aaa": [1, 1, 2, 2], "bbb": ["a", "b", "c", "d"]})
                )
                df = df.partition(by=["aaa"]).take(1)
                df.show()
            dag.run(self.engine)
            # Partition by and presort with NULLs
            # Column c needs to be kept even if not in presort or partition
            with FugueWorkflow() as dag:
                a = dag.df(
                    [
                        ["0", 1, 1],
                        ["0", 2, 1],
                        ["1", 3, 1],
                        ["1", 4, 1],
                        [None, 2, 1],
                        [None, 3, 1],
                    ],
                    "a:str,b:int,c:long",
                )
                a.partition(by=["a"], presort="b desc").take(n=1).assert_eq(
                    ArrayDataFrame(
                        [["0", 2, 1], ["1", 4, 1], [None, 3, 1]],
                        "a:str,b:int,c:long",
                    )
                )
            dag.run(self.engine)
            # No partition
            with FugueWorkflow() as dag:
                a = dag.df([[0, 1], [0, 2], [1, 3]], "a:int,b:int")
                a.take(1, presort="a,b desc").assert_eq(
                    ArrayDataFrame([[0, 2]], "a:int,b:int")
                )
            dag.run(self.engine)
            # take presort overrides partition presort
            with FugueWorkflow() as dag:
                a = dag.df(
                    [["0", 1], ["0", 2], ["1", 3], ["1", 4], [None, 2], [None, 3]],
                    "a:str,b:double",
                )
                a.partition(by=["a"], presort="b desc").take(
                    n=1, presort="b asc"
                ).assert_eq(
                    ArrayDataFrame([["0", 1], ["1", 3], [None, 2]], "a:str,b:double")
                )
            dag.run(self.engine)
            # order by with NULL first
            with FugueWorkflow() as dag:
                a = dag.df([["0", 1], ["0", 2], [None, 3]], "a:str,b:double")
                a.take(1, presort="a desc", na_position="first").assert_eq(
                    ArrayDataFrame([[None, 3]], "a:str,b:double")
                )
            dag.run(self.engine)
            # order by with NULL last
            with FugueWorkflow() as dag:
                a = dag.df([["0", 1], ["1", 2], [None, 3]], "a:str,b:double")
                a.take(1, presort="a desc", na_position="last").assert_eq(
                    ArrayDataFrame([["1", 2]], "a:str,b:double")
                )
            dag.run(self.engine)
            # Return any row because no presort
            with FugueWorkflow() as dag:
                a = dag.df([[0, 1], [0, 2], [1, 3]], "a:int,b:int")
                a.take(1).show()
            dag.run(self.engine)
            # n must be int
            with FugueWorkflow() as dag:
                with raises(ValueError):
                    a = dag.df([[0, 1], [0, 2], [1, 3]], "a:int,b:int")
                    a.take(0.5).show()
            dag.run(self.engine)
            # na_position not valid
            with FugueWorkflow() as dag:
                with raises(ValueError):
                    a = dag.df([[0, 1], [0, 2], [1, 3]], "a:int,b:int")
                    a.take(1, na_position=["True", "False"]).show()
            dag.run(self.engine)
            # Multiple takes
            with FugueWorkflow() as dag:
                df = dag.df(
                    pd.DataFrame({"aaa": [1, 1, 2, 2], "bbb": ["a", "b", "c", "d"]})
                )
                df1 = (
                    df.partition(by=["aaa"], presort="bbb")
                    .take(1)
                    .take(1, presort="aaa")
                )
                df1.show()
                df2 = (
                    df.partition(by=["aaa"], presort="bbb")
                    .take(1)
                    .partition(by=["aaa"], presort="bbb")
                    .take(1, presort="aaa")
                )
                df2.show()
            dag.run(self.engine)

        def test_col_ops(self):
            with FugueWorkflow() as dag:
                a = dag.df([[1, 10], [2, 20]], "x:long,y:long")
                aa = dag.df([[1, 10], [2, 20]], "xx:long,y:long")
                a.rename({"x": "xx"}).assert_eq(aa)
                a[["x"]].assert_eq(ArrayDataFrame([[1], [2]], "x:long"))

                a.drop(["y", "yy"], if_exists=True).assert_eq(
                    ArrayDataFrame([[1], [2]], "x:long")
                )

                a[["x"]].rename(x="xx").assert_eq(ArrayDataFrame([[1], [2]], "xx:long"))
                a.alter_columns("x:str").assert_eq(
                    ArrayDataFrame([["1", 10], ["2", 20]], "x:str,y:long")
                )
            dag.run(self.engine)

        def test_datetime_in_workflow(self):
            # schema: a:date,b:datetime
            def t1(df: pd.DataFrame) -> pd.DataFrame:
                df["b"] = "2020-01-02"
                df["b"] = pd.to_datetime(df["b"])
                return df

            class T2(Transformer):
                def get_output_schema(self, df):
                    return df.schema

                def transform(self, df):
                    # test for issue https://github.com/fugue-project/fugue/issues/92
                    return PandasDataFrame(df.as_pandas())

            with FugueWorkflow() as dag:
                a = dag.df([["2020-01-01"]], "a:date").transform(t1)
                b = dag.df(
                    [[datetime.date(2020, 1, 1), datetime.datetime(2020, 1, 2)]],
                    "a:date,b:datetime",
                )
                b.assert_eq(a, no_pandas=True)
                c = dag.df([["2020-01-01", "2020-01-01 00:00:00"]], "a:date,b:datetime")
                c.transform(T2).assert_eq(c)
                c.partition(by=["a"]).transform(T2).assert_eq(c)
            dag.run(self.engine)

        @pytest.fixture(autouse=True)
        def init_tmpdir(self, tmpdir):
            self.tmpdir = tmpdir

        def test_io(self):
            path = os.path.join(self.tmpdir, "a")
            path2 = os.path.join(self.tmpdir, "b.test.csv")
            path3 = os.path.join(self.tmpdir, "c.partition")
            with FugueWorkflow() as dag:
                b = dag.df([[6, 1], [2, 7]], "c:int,a:long")
                b.partition(num=3).save(path, fmt="parquet", single=True)
                b.save(path2, header=True)
            dag.run(self.engine)
            assert _LOCAL_FS.isfile(path)
            with FugueWorkflow() as dag:
                a = dag.load(path, fmt="parquet", columns=["a", "c"])
                a.assert_eq(dag.df([[1, 6], [7, 2]], "a:long,c:int"))
                a = dag.load(path2, header=True, columns="c:int,a:long")
                a.assert_eq(dag.df([[6, 1], [2, 7]], "c:int,a:long"))
            dag.run(self.engine)
            with FugueWorkflow() as dag:
                b = dag.df([[6, 1], [2, 7]], "c:int,a:long")
                b.partition(by="c").save(path3, fmt="parquet", single=False)
            dag.run(self.engine)
            assert _LOCAL_FS.isdir(path3)
            assert _LOCAL_FS.isdir(os.path.join(path3, "c=6"))
            assert _LOCAL_FS.isdir(os.path.join(path3, "c=2"))
            # TODO: in test below, once issue #288 is fixed, use dag.load
            #  instead of pd.read_parquet
            pdf = pd.read_parquet(path3).sort_values("a").reset_index(drop=True)
            pdf["c"] = pdf["c"].astype(int)
            pd.testing.assert_frame_equal(
                pdf,
                pd.DataFrame({"c": [6, 2], "a": [1, 7]}).reset_index(drop=True),
                check_like=True,
                check_dtype=False,
            )

        def test_save_and_use(self):
            path = os.path.join(self.tmpdir, "a")
            with FugueWorkflow() as dag:
                b = dag.df([[6, 1], [2, 7]], "c:int,a:long")
                c = b.save_and_use(path, fmt="parquet")
                b.assert_eq(c)
            dag.run(self.engine)

            with FugueWorkflow() as dag:
                b = dag.df([[6, 1], [2, 7]], "c:int,a:long")
                d = dag.load(path, fmt="parquet")
                b.assert_eq(d)
            dag.run(self.engine)

        def test_transformer_validation(self):
            # partitionby_has: b
            # input_has: a
            # schema: *
            def t1(df: pd.DataFrame) -> pd.DataFrame:
                return df

            @transformer("*", partitionby_has=["b"], input_has=["a"])
            def t2(df: pd.DataFrame) -> pd.DataFrame:
                return df

            class T3(Transformer):
                @property
                def validation_rules(self):
                    return dict(partitionby_has=["b"], input_has=["a"])

                def get_output_schema(self, df: DataFrame) -> Any:
                    return df.schema

                def transform(self, df: DataFrame) -> LocalDataFrame:
                    return df.as_local()

            for t in [t1, t2, T3]:
                # compile time
                with raises(FugueWorkflowCompileValidationError):
                    FugueWorkflow().df([[0, 1]], "a:int,b:int").transform(t)

                # runtime
                with raises(FugueWorkflowRuntimeValidationError):
                    with FugueWorkflow() as dag:
                        dag.df([[0, 1]], "c:int,b:int").partition(by=["b"]).transform(t)
                    dag.run(self.engine)

                with FugueWorkflow() as dag:
                    dag.df([[0, 1]], "a:int,b:int").partition(by=["b"]).transform(
                        t
                    ).assert_eq(dag.df([[0, 1]], "a:int,b:int"))
                dag.run(self.engine)

        def test_processor_validation(self):
            # input_has: a
            def p1(dfs: DataFrames) -> DataFrame:
                return dfs[0]

            @processor(input_has=["a"])
            def p2(dfs: DataFrames) -> DataFrame:
                return dfs[0]

            class P3(Processor):
                @property
                def validation_rules(self):
                    return dict(input_has=["a"])

                def process(self, dfs: DataFrames) -> DataFrame:
                    return dfs[0]

            for p in [p1, p2, P3]:
                # run time
                with raises(FugueWorkflowRuntimeValidationError):
                    with FugueWorkflow() as dag:
                        df1 = dag.df([[0, 1]], "a:int,b:int")
                        df2 = dag.df([[0, 1]], "c:int,d:int")
                        dag.process([df1, df2], using=p)
                    dag.run(self.engine)

                with FugueWorkflow() as dag:
                    df1 = dag.df([[0, 1]], "a:int,b:int")
                    df2 = dag.df([[0, 1]], "a:int,b:int")
                    dag.process([df1, df2], using=p).assert_eq(df1)
                dag.run(self.engine)

            # input_has: a
            # partitionby_has: b
            def p4(dfs: DataFrames) -> DataFrame:
                return dfs[0]

            # compile time
            with raises(FugueWorkflowCompileValidationError):
                dag = FugueWorkflow()
                df = dag.df([[0, 1]], "a:int,b:int")
                df.process(p4)

            with FugueWorkflow() as dag:
                dag.df([[0, 1]], "a:int,b:int").partition(by=["b"]).process(p4)
            dag.run(self.engine)

        def test_outputter_validation(self):
            # input_has: a
            def o1(dfs: DataFrames) -> None:
                pass

            @outputter(input_has=["a"])
            def o2(dfs: DataFrames) -> None:
                pass

            class O3(Outputter):
                @property
                def validation_rules(self):
                    return dict(input_has=["a"])

                def process(self, dfs: DataFrames) -> None:
                    pass

            for o in [o1, o2, O3]:
                # run time
                with raises(FugueWorkflowRuntimeValidationError):
                    with FugueWorkflow() as dag:
                        df1 = dag.df([[0, 1]], "a:int,b:int")
                        df2 = dag.df([[0, 1]], "c:int,d:int")
                        dag.output([df1, df2], using=o)
                    dag.run(self.engine)

                with FugueWorkflow() as dag:
                    df1 = dag.df([[0, 1]], "a:int,b:int")
                    df2 = dag.df([[0, 1]], "a:int,b:int")
                    dag.output([df1, df2], using=o)
                dag.run(self.engine)

            # input_has: a
            # partitionby_has: b
            def o4(dfs: DataFrames) -> None:
                pass

            # compile time
            with raises(FugueWorkflowCompileValidationError):
                dag = FugueWorkflow()
                df = dag.df([[0, 1]], "a:int,b:int")
                df.output(o4)

            with FugueWorkflow() as dag:
                dag.df([[0, 1]], "a:int,b:int").partition(by=["b"]).output(o4)
            dag.run(self.engine)

        def test_extension_registry(self):
            def my_creator() -> pd.DataFrame:
                return pd.DataFrame([[0, 1], [1, 2]], columns=["a", "b"])

            def my_processor(df: pd.DataFrame) -> pd.DataFrame:
                return df

            # schema: *
            def my_transformer(df: pd.DataFrame) -> pd.DataFrame:
                return df

            def my_out_transformer(df: pd.DataFrame) -> None:
                print(df)

            def my_outputter(df: pd.DataFrame) -> None:
                print(df)

            register_creator("mc", my_creator)
            register_processor("mp", my_processor)
            register_transformer("mt", my_transformer)
            register_output_transformer("mot", my_out_transformer)
            register_outputter("mo", my_outputter)

            with FugueWorkflow() as dag:
                df = dag.create("mc").process("mp").transform("mt")
                df.out_transform("mot")
                df.output("mo")
            dag.run(self.engine)

        def test_callback(self):  # noqa: C901
            class Callbacks(object):
                def __init__(self):
                    self.n = 0
                    self._lock = SerializableRLock()

                def call(self, value: int) -> int:
                    with self._lock:
                        self.n += value
                        return self.n

            cb = Callbacks()

            class CallbackTransformer(Transformer):
                def get_output_schema(self, df):
                    return df.schema

                def transform(self, df):
                    has = self.params.get_or_throw("has", bool)
                    v = self.cursor.key_value_array[0]
                    assert self.has_callback == has
                    if self.has_callback:
                        print(self.callback(v))
                    return df

            with FugueWorkflow() as dag:
                df = dag.df([[1, 1], [1, 2], [2, 3], [5, 6]], "a:int,b:int")
                res = df.partition(by=["a"]).transform(
                    CallbackTransformer, callback=cb.call, params=dict(has=True)
                )
                df.assert_eq(res)
                res = df.partition(by=["a"]).transform(
                    CallbackTransformer, params=dict(has=False)
                )
                df.assert_eq(res)
            dag.run(self.engine)

            assert 8 == cb.n

            # interfaceless tests
            cb2 = Callbacks()

            # schema: *
            def t0(df: pd.DataFrame) -> pd.DataFrame:
                return df

            # schema: *
            def t1(df: pd.DataFrame, c: Callable[[int], int]) -> pd.DataFrame:
                c(1)
                return df

            # schema: *
            def t12(
                df: pd.DataFrame, c: Optional[Callable[[int], int]] = None
            ) -> pd.DataFrame:
                if c is not None:
                    c(1)
                return df

            def t2(df: pd.DataFrame, c: Callable[[int], int]) -> None:
                c(1)

            with FugueWorkflow() as dag:
                df = dag.df([[1, 1], [1, 2], [2, 3], [5, 6]], "a:int,b:int")
                df.partition(by=["a"]).transform(
                    t0, callback=cb2.call
                ).persist()  # no effect because t0 ignores callback
                res = df.partition(by=["a"]).transform(t1, callback=cb2.call)  # +3
                df.partition(by=["a"]).out_transform(t2, callback=cb2.call)  # +3
                df.partition(by=["a"]).out_transform(t12, callback=cb2.call)  # +3
                df.partition(by=["a"]).out_transform(t12)
                raises(
                    FugueInterfacelessError,
                    lambda: (df.partition(by=["a"]).out_transform(t1)),
                )  # for t1, callback must be provided
                df.assert_eq(res)
            dag.run(self.engine)

            assert 9 == cb2.n

            cb2 = Callbacks()

            # schema: a:int,b:int
            def t3(
                df1: pd.DataFrame, df2: pd.DataFrame, c: Callable[[int], int]
            ) -> pd.DataFrame:
                c(1)
                return df1

            def t4(
                df1: pd.DataFrame, df2: pd.DataFrame, c: Callable[[int], int]
            ) -> None:
                c(1)

            def t42(
                df1: pd.DataFrame, df2: pd.DataFrame, c: Optional[Callable[[int], int]]
            ) -> None:
                if c is not None:
                    c(1)

            with FugueWorkflow() as dag:
                df1 = dag.df([[1, 1], [1, 2], [2, 3], [5, 6]], "a:int,b:int")
                df2 = dag.df([[1, 1], [1, 2], [2, 3], [5, 6]], "a:int,c:int")
                res = df1.zip(df2).transform(t3, callback=cb2.call).persist()  # +3
                df1.zip(df2).out_transform(t4, callback=cb2.call)  # +3
                df1.zip(df2).out_transform(t42, callback=cb2.call)  # +3
                df1.zip(df2).out_transform(t42)
                raises(
                    FugueInterfacelessError,
                    lambda: df1.zip(df2).transform(t3).persist(),
                )  # for t3, callback must be provided
            dag.run(self.engine)

            assert 9 == cb2.n

            def t5(df: pd.DataFrame, c: Callable) -> List[List[Any]]:
                c(df.shape[0])
                return [[pickle.dumps(123)]]

            cb3 = Callbacks()

            with FugueWorkflow() as dag:
                df = dag.df([[0], [1], [2], [3]], "a:int")
                df = df.transform(t5, schema="a:binary", callback=cb3.call)
                df.persist().yield_dataframe_as("x", as_local=True)
            dag.run(self.engine)
            assert dag.yields["x"].result.is_local

            assert 4 == cb3.n

        @pytest.mark.skipif(not HAS_QPD, reason="qpd not working")
        def test_sql_api(self):
            def tr(df: pd.DataFrame, n=1) -> pd.DataFrame:
                return df + n

            with fa.engine_context(self.engine):
                df1 = fa.as_fugue_df([[0, 1], [2, 3], [4, 5]], schema="a:long,b:int")
                df2 = pd.DataFrame([[0, 10], [1, 100]], columns=["a", "c"])
                sdf1 = fa.raw_sql(  # noqa
                    "SELECT ", df1, ".a, b FROM ", df1, " WHERE a<4"
                )
                sdf2 = fa.raw_sql("SELECT * FROM ", df2, " WHERE a<1")  # noqa

                sdf3 = fa.fugue_sql(
                    """
                SELECT sdf1.a,sdf1.b,c FROM sdf1 INNER JOIN sdf2 ON sdf1.a=sdf2.a
                TRANSFORM USING tr SCHEMA *
                """
                )
                res = fa.fugue_sql_flow(
                    """
                TRANSFORM x USING tr(n=2) SCHEMA *
                YIELD LOCAL DATAFRAME AS res
                PRINT sdf1
                """,
                    x=sdf3,
                ).run()
                df_eq(
                    res["res"],
                    [[3, 4, 13]],
                    schema="a:long,b:int,c:long",
                    check_schema=False,
                    throw=True,
                )

                sdf4 = fa.fugue_sql(
                    """
                SELECT sdf1.a,b,c FROM sdf1 INNER JOIN sdf2 ON sdf1.a=sdf2.a
                TRANSFORM USING tr SCHEMA *
                """,
                    as_fugue=False,
                    as_local=True,
                )
                assert not isinstance(sdf4, DataFrame)
                assert fa.is_local(sdf4)

        @pytest.mark.skipif(os.name == "nt", reason="Skip Windows")
        @pytest.mark.skipif(not HAS_QPD, reason="qpd not working")
        def test_any_column_name(self):
            f_parquet = os.path.join(str(self.tmpdir), "a.parquet")
            f_csv = os.path.join(str(self.tmpdir), "a.csv")

            # schema: *,`c *`:long
            def tr(df: pd.DataFrame) -> pd.DataFrame:
                return df.assign(**{"c *": 2})

            with fa.engine_context(self.engine):
                df1 = pd.DataFrame([[0, 1], [2, 3]], columns=["a b", " "])
                df2 = pd.DataFrame([[0, 10], [20, 3]], columns=["a b", "d"])
                r = fa.inner_join(df1, df2, as_fugue=True)
                df_eq(r, [[0, 1, 10]], "`a b`:long,` `:long,d:long", throw=True)
                r = fa.transform(r, tr)
                df_eq(
                    r,
                    [[0, 1, 10, 2]],
                    "`a b`:long,` `:long,d:long,`c *`:long",
                    throw=True,
                )
                r = fa.alter_columns(r, "`c *`:str")
                r = fa.select(
                    r,
                    col("a b").alias("a b "),
                    col(" ").alias("x y"),
                    col("d"),
                    col("c *").cast(int),
                )
                df_eq(
                    r,
                    [[0, 1, 10, 2]],
                    "`a b `:long,`x y`:long,d:long,`c *`:long",
                    throw=True,
                )
                r = fa.rename(r, {"a b ": "a b"})
                fa.save(r, f_csv, header=True, force_single=True)
                fa.save(r, f_parquet)
                df_eq(
                    fa.load(f_parquet, columns=["x y", "d", "c *"], as_fugue=True),
                    [[1, 10, 2]],
                    "`x y`:long,d:long,`c *`:long",
                    throw=True,
                )
                df_eq(
                    fa.load(
                        f_csv,
                        header=True,
                        infer_schema=False,
                        columns=["d", "c *"],
                        as_fugue=True,
                    ),
                    [["10", "2"]],
                    "d:str,`c *`:str",
                    throw=True,
                )
                df_eq(
                    fa.load(
                        f_csv,
                        header=True,
                        columns="`a b`:long,`x y`:long,d:long,`c *`:long",
                        as_fugue=True,
                    ),
                    [[0, 1, 10, 2]],
                    "`a b`:long,`x y`:long,d:long,`c *`:long",
                    throw=True,
                )

                r = fa.fugue_sql(
                    """
                df1 = CREATE [[0, 1], [2, 3]] SCHEMA `a b`:long,` `:long
                df2 = CREATE [[0, 10], [20, 3]] SCHEMA `a b`:long,d:long
                SELECT df1.*,d FROM df1 INNER JOIN df2 ON df1.`a b`=df2.`a b`
                """,
                    as_fugue=True,
                )
                df_eq(r, [[0, 1, 10]], "`a b`:long,` `:long,d:long", throw=True)
                r = fa.fugue_sql(
                    """
                TRANSFORM r USING tr SCHEMA *,`c *`:long
                """,
                    as_fugue=True,
                )
                df_eq(
                    r,
                    [[0, 1, 10, 2]],
                    "`a b`:long,` `:long,d:long,`c *`:long",
                    throw=True,
                )
                r = fa.fugue_sql(
                    """
                ALTER COLUMNS `c *`:long FROM r
                """,
                    as_fugue=True,
                )
                df_eq(
                    r,
                    [[0, 1, 10, 2]],
                    "`a b`:long,` `:long,d:long,`c *`:long",
                    throw=True,
                )
                res = fa.fugue_sql_flow(
                    """
                LOAD "{{f_parquet}}" COLUMNS `x y`,d,`c *`
                YIELD LOCAL DATAFRAME AS r1

                LOAD "{{f_csv}}"(header=TRUE,infer_schema=FALSE) COLUMNS `x y`,d,`c *`
                YIELD LOCAL DATAFRAME AS r2

                LOAD "{{f_csv}}"(header=TRUE,infer_schema=FALSE)
                COLUMNS `a b`:long,`x y`:long,d:long,`c *`:long
                YIELD LOCAL DATAFRAME AS r3
                """,
                    f_parquet=f_parquet,
                    f_csv=f_csv,
                ).run()
                df_eq(
                    res["r1"],
                    [[1, 10, 2]],
                    "`x y`:long,d:long,`c *`:long",
                    throw=True,
                )
                df_eq(
                    res["r2"],
                    [["1", "10", "2"]],
                    "`x y`:str,d:str,`c *`:str",
                    throw=True,
                )
                df_eq(
                    res["r3"],
                    [[0, 1, 10, 2]],
                    "`a b`:long,`x y`:long,d:long,`c *`:long",
                    throw=True,
                )


def mock_creator(p: int) -> DataFrame:
    return ArrayDataFrame([[p]], "a:int")


def mock_creator2(p: int) -> AnyDataFrame:
    return fa.as_fugue_df([[p]], schema="a:int")


def mock_processor(df1: List[List[Any]], df2: List[List[Any]]) -> DataFrame:
    return ArrayDataFrame([[len(df1) + len(df2)]], "a:int")


def mock_processor2(e: ExecutionEngine, dfs: DataFrames) -> DataFrame:
    assert "test" in e.conf
    return ArrayDataFrame([[sum(s.count() for s in dfs.values())]], "a:int")


class MockProcessor3(Processor):
    def process(self, dfs):
        assert "test" in self.workflow_conf
        return ArrayDataFrame([[sum(s.count() for s in dfs.values())]], "a:int")


def mock_processor4(df1: AnyDataFrame, df2: AnyDataFrame) -> AnyDataFrame:
    return ArrayDataFrame([[fa.count(df1) + fa.count(df2)]], "a:int")


def mock_outputter(df1: List[List[Any]], df2: List[List[Any]]) -> None:
    assert len(df1) == len(df2)


def mock_outputter2(df: List[List[Any]]) -> None:
    print(df)


class MockOutputter3(Outputter):
    def process(self, dfs):
        assert "3" == self.partition_spec.num_partitions


def mock_outputter4(df1: AnyDataFrame, df2: AnyDataFrame) -> None:
    assert fa.count(df1) == fa.count(df2)


class MockOutputter4(Outputter):
    def process(self, dfs):
        for k, v in dfs.items():
            print(k)
            v.show()


class MockTransform1(Transformer):
    def get_output_schema(self, df: DataFrame) -> Any:
        assert "test" in self.workflow_conf
        return [df.schema, "ct:int,p:int"]

    def on_init(self, df: DataFrame) -> None:
        assert "test" in self.workflow_conf
        self.pn = self.cursor.physical_partition_no
        self.ks = self.key_schema
        if "on_init_called" not in self.__dict__:
            self.on_init_called = 1
        else:  # pragma: no cover
            self.on_init_called += 1

    def transform(self, df: LocalDataFrame) -> LocalDataFrame:
        assert 1 == self.on_init_called
        assert "test" in self.workflow_conf
        pdf = df.as_pandas()
        pdf["p"] = self.params.get("p", 1)
        pdf["ct"] = pdf.shape[0]
        return PandasDataFrame(pdf, self.output_schema)


@transformer(lambda df, **kwargs: df.schema + (kwargs.get("col", "p") + ":int"))
def mock_tf0(df: pd.DataFrame, p=1, col="p") -> pd.DataFrame:
    df[col] = p
    return df


# schema: *,ct:int,p:int
def mock_tf1(df: pd.DataFrame, p=1) -> pd.DataFrame:
    df["p"] = p
    df["ct"] = df.shape[0]
    return df


def mock_tf2_except(df: Iterable[Dict[str, Any]], p=1) -> Iterable[Dict[str, Any]]:
    n = 0
    for row in df:
        yield row
        n += 1
        if n > 1:
            raise NotImplementedError


# schema: *
def mock_tf3(df: Iterable[Dict[str, Any]]) -> Iterable[Dict[str, Any]]:
    # binary data test
    for row in df:
        obj = pickle.loads(row["b"])
        obj[0] += 1
        obj[1] += "x"
        row["b"] = pickle.dumps(obj)
        yield row


class MockCoTransform1(CoTransformer):
    def get_output_schema(self, dfs: DataFrames) -> Any:
        assert "test" in self.workflow_conf
        assert 2 == len(dfs)
        if self.params.get("named", False):
            assert dfs.has_key
        else:
            assert not dfs.has_key
        return [self.key_schema, "ct1:int,ct2:int,p:int"]

    def on_init(self, dfs: DataFrames) -> None:
        assert "test" in self.workflow_conf
        assert 2 == len(dfs)
        if self.params.get("named", False):
            assert dfs.has_key
        else:
            assert not dfs.has_key
        self.pn = self.cursor.physical_partition_no
        self.ks = self.key_schema
        if "on_init_called" not in self.__dict__:
            self.on_init_called = 1
        else:  # pragma: no cover
            self.on_init_called += 1

    def transform(self, dfs: DataFrames) -> LocalDataFrame:
        assert 1 == self.on_init_called
        assert "test" in self.workflow_conf
        assert 2 == len(dfs)
        if self.params.get("named", False):
            assert dfs.has_key
        else:
            assert not dfs.has_key
        row = self.cursor.key_value_array + [
            dfs[0].count(),
            dfs[1].count(),
            self.params.get("p", 1),
        ]
        return ArrayDataFrame([row], self.output_schema)


@cotransformer(
    lambda dfs, **kwargs: "a:int,ct1:int,ct2:int," + kwargs.get("col", "p") + ":int"
)
def mock_co_tf1(
    df1: List[Dict[str, Any]], df2: List[List[Any]], p=1, col="p"
) -> List[List[Any]]:
    return [[df1[0]["a"], len(df1), len(df2), p]]


def mock_co_tf2(dfs: DataFrames, p=1) -> List[List[Any]]:
    return [[dfs[0].peek_dict()["a"], dfs[0].count(), dfs[1].count(), p]]


@cotransformer(Schema("a:int,ct1:int,p:int"))
def mock_co_tf3(df1: List[Dict[str, Any]], p=1) -> List[List[Any]]:
    return [[df1[0]["a"], len(df1), p]]


@cotransformer("a:int,ct1:int,p:int")
def mock_co_tf4_ex(df1: List[Dict[str, Any]], p=1) -> List[List[Any]]:
    k = df1[0]["a"]
    if k == 2:
        raise NotImplementedError
    return [[df1[0]["a"], len(df1), p]]



from fugue.execution.factory import register_execution_engine, register_sql_engine
from fugue.execution.native_execution_engine import (
    NativeExecutionEngine,
    QPDPandasEngine,
)


def _register() -> None:
    """Register Fugue core additional types

    .. note::

        This function is automatically called when you do

        >>> import fugue
    """
    _register_engines()


def _register_engines() -> None:
    register_execution_engine(
        "native", lambda conf: NativeExecutionEngine(conf), on_dup="ignore"
    )
    register_execution_engine(
        "pandas", lambda conf: NativeExecutionEngine(conf), on_dup="ignore"
    )
    register_sql_engine(
        "qpdpandas", lambda engine: QPDPandasEngine(engine), on_dup="ignore"
    )
    register_sql_engine(
        "qpd_pandas", lambda engine: QPDPandasEngine(engine), on_dup="ignore"
    )



import pickle
from abc import ABC, abstractmethod
from types import LambdaType
from typing import Any, Callable, Dict
from uuid import uuid4

from triad import ParamDict, SerializableRLock, assert_or_throw, to_uuid
from triad.utils.convert import get_full_type_path, to_type


class RPCClient(object):
    """RPC client interface"""

    def __call__(self, *args: Any, **kwargs: Any) -> Any:  # pragma: no cover
        raise NotImplementedError


class RPCHandler(RPCClient):
    """RPC handler hosting the real logic on driver side"""

    def __init__(self):
        self._rpchandler_lock = SerializableRLock()
        self._running = 0

    @property
    def running(self) -> bool:
        """Whether the handler is in running state"""
        return self._running > 0

    def __uuid__(self) -> str:
        """UUID that can affect the determinism of the workflow"""
        return ""  # pragma: no cover

    def start_handler(self) -> None:
        """User implementation of starting the handler"""
        return

    def stop_handler(self) -> None:
        """User implementation of stopping the handler"""
        return

    def start(self) -> "RPCHandler":
        """Start the handler, wrapping :meth:`~.start_handler`

        :return: the instance itself
        """
        with self._rpchandler_lock:
            if self._running == 0:
                self.start_handler()
            self._running += 1
        return self

    def stop(self) -> None:
        """Stop the handler, wrapping :meth:`~.stop_handler`"""
        with self._rpchandler_lock:
            if self._running == 1:
                self.stop_handler()
            self._running -= 1
            if self._running < 0:
                self._running = 0

    def __enter__(self) -> "RPCHandler":
        """``with`` statement. :meth:`~.start` must be called

        .. admonition:: Examples

            .. code-block:: python

                with handler.start():
                    handler...
        """
        with self._rpchandler_lock:
            assert_or_throw(self._running, "use `with <instance>.start():` instead")
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.stop()

    def __getstate__(self):
        """
        :raises pickle.PicklingError: serialization
          of ``RPCHandler`` is not allowed
        """
        raise pickle.PicklingError(f"{self} is not serializable")

    def __copy__(self) -> "RPCHandler":
        """Copy takes no effect

        :return: the instance itself
        """
        return self

    def __deepcopy__(self, memo: Any) -> "RPCHandler":
        """Deep copy takes no effect

        :return: the instance itself
        """
        return self


class EmptyRPCHandler(RPCHandler):
    """The class representing empty :class:`~.RPCHandler`"""


class RPCServer(RPCHandler, ABC):
    """Server abstract class hosting multiple :class:`~.RPCHandler`.

    :param conf: |FugueConfig|
    """

    def __init__(self, conf: Any):
        super().__init__()
        self._conf = ParamDict(conf)
        self._handlers: Dict[str, RPCHandler] = {}

    @property
    def conf(self) -> ParamDict:
        """Config initialized this instance"""
        return self._conf

    @abstractmethod
    def make_client(self, handler: Any) -> RPCClient:
        """Make a :class:`~.RPCHandler` and return the correspondent
        :class:`~.RPCClient`

        :param handler: |RPCHandlerLikeObject|
        :return: the client connecting to the handler
        """
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def start_server(self) -> None:
        """User implementation of starting the server"""
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def stop_server(self) -> None:
        """User implementation of stopping the server"""
        raise NotImplementedError  # pragma: no cover

    def start_handler(self) -> None:
        """Wrapper to start the server, do not override or call directly"""
        with self._rpchandler_lock:
            self.start_server()

    def stop_handler(self) -> None:
        """Wrapper to stop the server, do not override or call directly"""
        with self._rpchandler_lock:
            self.stop_server()
            for v in self._handlers.values():
                if v.running:
                    v.stop()
            self._handlers.clear()

    def invoke(self, key: str, *args: Any, **kwargs: Any) -> Any:
        """Invoke the correspondent handler

        :param key: key of the handler
        :return: the return value of the handler
        """
        with self._rpchandler_lock:
            handler = self._handlers[key]
        return handler(*args, **kwargs)

    def register(self, handler: Any) -> str:
        """Register the hander into the server

        :param handler: |RPCHandlerLikeObject|
        :return: the unique key of the handler
        """
        with self._rpchandler_lock:
            key = "_" + str(uuid4()).split("-")[-1]
            assert_or_throw(key not in self._handlers, f"{key} already exists")
            self._handlers[key] = to_rpc_handler(handler).start()
            return key


class NativeRPCClient(RPCClient):
    """Native RPC Client that can only be used locally.
    Use :meth:`~.NativeRPCServer.make_client` to create this instance.

    :param server: the parent :class:`~.NativeRPCServer`
    :param key: the unique key for the handler and this client
    """

    def __init__(self, server: "NativeRPCServer", key: str):
        self._key = key
        self._server = server

    def __call__(self, *args: Any, **kwargs: Any) -> str:
        return self._server.invoke(self._key, *args, **kwargs)

    def __getstate__(self):
        raise pickle.PicklingError(f"{self} is not serializable")


class NativeRPCServer(RPCServer):
    """Native RPC Server that can only be used locally.

    :param conf: |FugueConfig|
    """

    def make_client(self, handler: Any) -> RPCClient:
        """Add ``handler`` and correspondent :class:`~.NativeRPCClient`

        :param handler: |RPCHandlerLikeObject|
        :return: the native RPC client
        """
        key = self.register(handler)
        return NativeRPCClient(self, key)

    def start_server(self) -> None:
        """Do nothing"""
        return

    def stop_server(self) -> None:
        """Do nothing"""
        return


class RPCFunc(RPCHandler):
    """RPCHandler wrapping a python function.

    :param func: a python function
    """

    def __init__(self, func: Callable):
        super().__init__()
        assert_or_throw(callable(func), lambda: ValueError(func))
        self._func = func
        if isinstance(func, LambdaType):
            self._uuid = to_uuid("lambda")
        else:
            self._uuid = to_uuid(get_full_type_path(func))

    def __uuid__(self) -> str:
        """If the underlying function is a static function, then the full
        type path of the function determines the uuid, but for a lambda
        function, the uuid is a constant, so it could be overly
        deterministic

        :return: the unique id
        """
        return self._uuid

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self._func(*args, **kwargs)


def to_rpc_handler(obj: Any) -> RPCHandler:
    """Convert object to :class:`~.RPCHandler`. If the object is already
    ``RPCHandler``, then the original instance will be returned.
    If the object is ``None`` then :class:`~.EmptyRPCHandler` will be returned.
    If the object is a python function then :class:`~.RPCFunc` will be returned.

    :param obj: |RPCHandlerLikeObject|
    :return: the RPC handler
    """
    if obj is None:
        return EmptyRPCHandler()
    if isinstance(obj, RPCHandler):
        return obj
    if callable(obj):
        return RPCFunc(obj)
    raise ValueError(obj)


def make_rpc_server(conf: Any) -> RPCServer:
    """Make :class:`~.RPCServer` based on configuration.
    If '`fugue.rpc.server`` is set, then the value will be used as
    the server type for the initialization. Otherwise, a
    :class:`~.NativeRPCServer` instance will be returned

    :param conf: |FugueConfig|
    :return: the RPC server
    """
    conf = ParamDict(conf)
    tp = conf.get_or_none("fugue.rpc.server", str)
    t_server = NativeRPCServer if tp is None else to_type(tp, RPCServer)
    return t_server(conf)  # type: ignore



from typing import Any, Dict, Tuple

from triad.utils.assertion import assert_or_throw
from triad.utils.convert import get_caller_global_local_vars

from fugue._utils.misc import import_fsql_dependency

from ..collections.yielded import Yielded
from ..constants import FUGUE_CONF_SQL_DIALECT, FUGUE_CONF_SQL_IGNORE_CASE
from ..dataframe.api import is_df
from ..dataframe.dataframe import DataFrame
from ..workflow.workflow import FugueWorkflow, WorkflowDataFrame, WorkflowDataFrames
from ._utils import LazyWorkflowDataFrame, fill_sql_template


class FugueSQLWorkflow(FugueWorkflow):
    """Fugue workflow that supports Fugue SQL. Please read |FugueSQLTutorial|."""

    def __init__(self, compile_conf: Any = None):
        super().__init__(compile_conf)
        self._sql_vars: Dict[str, WorkflowDataFrame] = {}

    @property
    def sql_vars(self) -> Dict[str, WorkflowDataFrame]:
        return self._sql_vars

    def __call__(self, code: str, *args: Any, **kwargs: Any) -> None:
        global_vars, local_vars = get_caller_global_local_vars()
        variables = self._sql(
            code, self._sql_vars, global_vars, local_vars, *args, **kwargs
        )
        for k, v in variables.items():
            if isinstance(v, WorkflowDataFrame) and v.workflow is self:
                self._sql_vars[k] = v

    def _sql(
        self, code: str, *args: Any, **kwargs: Any
    ) -> Dict[str, Tuple[WorkflowDataFrame, WorkflowDataFrames, LazyWorkflowDataFrame]]:
        FugueSQLParser = import_fsql_dependency("fugue_sql_antlr").FugueSQLParser

        from ._visitors import FugueSQLHooks, _Extensions

        # TODO: move dict construction to triad
        params: Dict[str, Any] = {}
        for a in args:
            assert_or_throw(
                isinstance(a, Dict), lambda: f"args can only have dict: {a}"
            )
            params.update(a)
        params.update(kwargs)
        params, dfs = self._split_params(params)
        code = fill_sql_template(code, params)
        sql = FugueSQLParser(
            code,
            "fugueLanguage",
            ignore_case=self.conf.get_or_throw(FUGUE_CONF_SQL_IGNORE_CASE, bool),
            parse_mode="auto",
        )
        v = _Extensions(
            sql,
            FugueSQLHooks(),
            self,
            dialect=self.conf.get_or_throw(FUGUE_CONF_SQL_DIALECT, str),
            variables=dfs,  # type: ignore
            local_vars=params,
        )
        v.visit(sql.tree)
        return v.variables

    def _split_params(
        self, params: Dict[str, Any]
    ) -> Tuple[Dict[str, Any], Dict[str, LazyWorkflowDataFrame]]:
        p: Dict[str, Any] = {}
        dfs: Dict[str, LazyWorkflowDataFrame] = {}
        for k, v in params.items():
            if isinstance(v, (int, str, float, bool)):
                p[k] = v
            elif isinstance(v, (DataFrame, Yielded)) or is_df(v):
                dfs[k] = LazyWorkflowDataFrame(k, v, self)
            else:
                p[k] = v
        return p, dfs



import json
from functools import partial
from typing import Any, Tuple

import matplotlib.pyplot as plt
import pandas as pd
import seaborn

from fugue import Outputter
from fugue.extensions import namespace_candidate, parse_outputter

from ..viz._ext import Visualize


@parse_outputter.candidate(namespace_candidate("sns", lambda x: isinstance(x, str)))
def _parse_seaborn(obj: Tuple[str, str]) -> Outputter:
    return _SeabornVisualize(obj[1])


class _SeabornVisualize(Visualize):
    def __init__(self, func: str) -> None:
        super().__init__(func)
        getattr(seaborn, func)  # ensure the func exists

    def _plot(self, df: pd.DataFrame) -> None:
        params = dict(self.params)
        title: Any = None
        if len(self.partition_spec.partition_by) > 0:
            keys = df[self.partition_spec.partition_by].head(1).to_dict("records")[0]
            kt = json.dumps(keys)[1:-1]
            if "title" in params:
                params["title"] = params["title"] + " -- " + kt
            else:
                params["title"] = kt
            df = df.drop(self.partition_spec.partition_by, axis=1)
        func = self._get_func(df)
        title = params.pop("title", None)
        plt.figure(0)
        func(**params).set(title=title)
        plt.show()

    def _get_func(self, df: pd.DataFrame) -> Any:
        f = getattr(seaborn, self._func)
        return partial(f, df)



from typing import Callable

from triad import conditional_dispatcher
from triad.utils.dispatcher import ConditionalDispatcher

from ..constants import FUGUE_ENTRYPOINT


def fugue_plugin(func: Callable) -> ConditionalDispatcher:
    return conditional_dispatcher(entry_point=FUGUE_ENTRYPOINT)(func)  # type: ignore



import inspect
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    no_type_check,
)

import pandas as pd
import pyarrow as pa
from triad import Schema, assert_or_throw
from triad.collections.function_wrapper import (
    AnnotatedParam,
    FunctionWrapper,
    KeywordParam,
    PositionalParam,
    function_wrapper,
)
from triad.utils.iter import EmptyAwareIterable, make_empty_aware

from ..constants import FUGUE_ENTRYPOINT
from ..dataset.api import count as df_count
from .array_dataframe import ArrayDataFrame
from .arrow_dataframe import ArrowDataFrame
from .dataframe import AnyDataFrame, DataFrame, LocalDataFrame, as_fugue_df
from .dataframe_iterable_dataframe import (
    IterableArrowDataFrame,
    IterablePandasDataFrame,
    LocalDataFrameIterableDataFrame,
)
from .dataframes import DataFrames
from .iterable_dataframe import IterableDataFrame
from .pandas_dataframe import PandasDataFrame


@function_wrapper(FUGUE_ENTRYPOINT)
class DataFrameFunctionWrapper(FunctionWrapper):
    @property
    def need_output_schema(self) -> Optional[bool]:
        return (
            self._rt.need_schema()
            if isinstance(self._rt, _DataFrameParamBase)
            else False
        )

    def get_format_hint(self) -> Optional[str]:
        for v in self._params.values():
            if isinstance(v, _DataFrameParamBase):
                if v.format_hint() is not None:
                    return v.format_hint()
        if isinstance(self._rt, _DataFrameParamBase):
            return self._rt.format_hint()
        return None

    def run(  # noqa: C901
        self,
        args: List[Any],
        kwargs: Dict[str, Any],
        ignore_unknown: bool = False,
        output_schema: Any = None,
        output: bool = True,
        ctx: Any = None,
    ) -> Any:
        p: Dict[str, Any] = {}
        for i in range(len(args)):
            p[self._params.get_key_by_index(i)] = args[i]
        p.update(kwargs)
        has_kw = False
        rargs: Dict[str, Any] = {}
        for k, v in self._params.items():
            if isinstance(v, (PositionalParam, KeywordParam)):
                if isinstance(v, KeywordParam):
                    has_kw = True
            elif k in p:
                if isinstance(v, _DataFrameParamBase):
                    assert_or_throw(
                        isinstance(p[k], DataFrame),
                        lambda: TypeError(f"{p[k]} is not a DataFrame"),
                    )
                    rargs[k] = v.to_input_data(p[k], ctx=ctx)
                else:
                    rargs[k] = p[k]  # TODO: should we do auto type conversion?
                del p[k]
            elif v.required:
                raise ValueError(f"{k} is required by not given")
        if has_kw:
            rargs.update(p)
        elif not ignore_unknown and len(p) > 0:
            raise ValueError(f"{p} are not acceptable parameters")
        rt = self._func(**rargs)
        if not output:
            if isinstance(self._rt, _DataFrameParamBase):
                self._rt.count(rt)
            return
        if isinstance(self._rt, _DataFrameParamBase):
            return self._rt.to_output_df(rt, output_schema, ctx=ctx)
        return rt


fugue_annotated_param = DataFrameFunctionWrapper.annotated_param


@fugue_annotated_param(
    "Callable",
    "F",
    lambda annotation: (
        annotation == Callable
        or annotation == callable  # pylint: disable=comparison-with-callable
        or str(annotation).startswith("typing.Callable")
    ),
)
class _CallableParam(AnnotatedParam):
    pass


@fugue_annotated_param(
    "Callable",
    "f",
    lambda annotation: (
        annotation == Optional[Callable]
        or annotation == Optional[callable]
        or str(annotation).startswith("typing.Union[typing.Callable")  # 3.8-
        or str(annotation).startswith("typing.Optional[typing.Callable")  # 3.9+
    ),
)
class _OptionalCallableParam(AnnotatedParam):
    pass


class _DataFrameParamBase(AnnotatedParam):
    def __init__(self, param: Optional[inspect.Parameter]):
        super().__init__(param)
        assert_or_throw(self.required, lambda: TypeError(f"{self} must be required"))

    def to_input_data(self, df: DataFrame, ctx: Any) -> Any:  # pragma: no cover
        raise NotImplementedError

    def to_output_df(
        self, df: Any, schema: Any, ctx: Any
    ) -> DataFrame:  # pragma: no cover
        raise NotImplementedError

    def count(self, df: Any) -> int:  # pragma: no cover
        raise NotImplementedError

    def need_schema(self) -> Optional[bool]:
        return False

    def format_hint(self) -> Optional[str]:
        return None


@fugue_annotated_param(DataFrame, "d", child_can_reuse_code=True)
class DataFrameParam(_DataFrameParamBase):
    def to_input_data(self, df: DataFrame, ctx: Any) -> Any:
        return df

    def to_output_df(self, output: Any, schema: Any, ctx: Any) -> DataFrame:
        assert_or_throw(
            schema is None or output.schema == schema,
            lambda: f"Output schema mismatch {output.schema} vs {schema}",
        )
        return output

    def count(self, df: Any) -> int:
        if df.is_bounded:
            return df.count()
        else:
            return sum(1 for _ in df.as_array_iterable())


@fugue_annotated_param(AnyDataFrame)
class _AnyDataFrameParam(DataFrameParam):
    def to_output_df(self, output: AnyDataFrame, schema: Any, ctx: Any) -> DataFrame:
        return (
            as_fugue_df(output)
            if schema is None
            else as_fugue_df(output, schema=schema)
        )

    def count(self, df: Any) -> int:
        return df_count(df)


@fugue_annotated_param(LocalDataFrame, "l", child_can_reuse_code=True)
class LocalDataFrameParam(DataFrameParam):
    def to_input_data(self, df: DataFrame, ctx: Any) -> LocalDataFrame:
        return df.as_local()

    def to_output_df(self, output: LocalDataFrame, schema: Any, ctx: Any) -> DataFrame:
        assert_or_throw(
            schema is None or output.schema == schema,
            lambda: f"Output schema mismatch {output.schema} vs {schema}",
        )
        return output

    def count(self, df: LocalDataFrame) -> int:
        if df.is_bounded:
            return df.count()
        else:
            return sum(1 for _ in df.as_array_iterable())


@fugue_annotated_param(
    "[NoSchema]", "s", matcher=lambda x: False, child_can_reuse_code=True
)
class _LocalNoSchemaDataFrameParam(LocalDataFrameParam):
    def need_schema(self) -> Optional[bool]:
        return True


@fugue_annotated_param(List[List[Any]])
class _ListListParam(_LocalNoSchemaDataFrameParam):
    @no_type_check
    def to_input_data(self, df: DataFrame, ctx: Any) -> List[List[Any]]:
        return df.as_array(type_safe=True)

    @no_type_check
    def to_output_df(self, output: List[List[Any]], schema: Any, ctx: Any) -> DataFrame:
        return ArrayDataFrame(output, schema)

    @no_type_check
    def count(self, df: List[List[Any]]) -> int:
        return len(df)


@fugue_annotated_param(
    Iterable[List[Any]],
    matcher=lambda x: x == Iterable[List[Any]] or x == Iterator[List[Any]],
)
class _IterableListParam(_LocalNoSchemaDataFrameParam):
    @no_type_check
    def to_input_data(self, df: DataFrame, ctx: Any) -> Iterable[List[Any]]:
        return df.as_array_iterable(type_safe=True)

    @no_type_check
    def to_output_df(
        self, output: Iterable[List[Any]], schema: Any, ctx: Any
    ) -> DataFrame:
        return IterableDataFrame(output, schema)

    @no_type_check
    def count(self, df: Iterable[List[Any]]) -> int:
        return sum(1 for _ in df)


@fugue_annotated_param(EmptyAwareIterable[List[Any]])
class _EmptyAwareIterableListParam(_LocalNoSchemaDataFrameParam):
    @no_type_check
    def to_input_data(self, df: DataFrame, ctx: Any) -> EmptyAwareIterable[List[Any]]:
        return make_empty_aware(df.as_array_iterable(type_safe=True))

    @no_type_check
    def to_output_df(
        self, output: EmptyAwareIterable[List[Any]], schema: Any, ctx: Any
    ) -> DataFrame:
        return IterableDataFrame(output, schema)

    @no_type_check
    def count(self, df: EmptyAwareIterable[List[Any]]) -> int:
        return sum(1 for _ in df)


@fugue_annotated_param(List[Dict[str, Any]])
class _ListDictParam(_LocalNoSchemaDataFrameParam):
    @no_type_check
    def to_input_data(self, df: DataFrame, ctx: Any) -> List[Dict[str, Any]]:
        return df.as_local().as_dicts()

    @no_type_check
    def to_output_df(
        self, output: List[Dict[str, Any]], schema: Any, ctx: Any
    ) -> DataFrame:
        schema = schema if isinstance(schema, Schema) else Schema(schema)

        def get_all() -> Iterable[List[Any]]:
            for row in output:
                yield [row[x] for x in schema.names]

        return IterableDataFrame(get_all(), schema)

    @no_type_check
    def count(self, df: List[Dict[str, Any]]) -> int:
        return len(df)


@fugue_annotated_param(
    Iterable[Dict[str, Any]],
    matcher=lambda x: x == Iterable[Dict[str, Any]] or x == Iterator[Dict[str, Any]],
)
class _IterableDictParam(_LocalNoSchemaDataFrameParam):
    @no_type_check
    def to_input_data(self, df: DataFrame, ctx: Any) -> Iterable[Dict[str, Any]]:
        return df.as_dict_iterable()

    @no_type_check
    def to_output_df(
        self, output: Iterable[Dict[str, Any]], schema: Any, ctx: Any
    ) -> DataFrame:
        schema = schema if isinstance(schema, Schema) else Schema(schema)

        def get_all() -> Iterable[List[Any]]:
            for row in output:
                yield [row[x] for x in schema.names]

        return IterableDataFrame(get_all(), schema)

    @no_type_check
    def count(self, df: Iterable[Dict[str, Any]]) -> int:
        return sum(1 for _ in df)


@fugue_annotated_param(EmptyAwareIterable[Dict[str, Any]])
class _EmptyAwareIterableDictParam(_LocalNoSchemaDataFrameParam):
    @no_type_check
    def to_input_data(
        self, df: DataFrame, ctx: Any
    ) -> EmptyAwareIterable[Dict[str, Any]]:
        return make_empty_aware(df.as_dict_iterable())

    @no_type_check
    def to_output_df(
        self, output: EmptyAwareIterable[Dict[str, Any]], schema: Any, ctx: Any
    ) -> DataFrame:
        schema = schema if isinstance(schema, Schema) else Schema(schema)

        def get_all() -> Iterable[List[Any]]:
            for row in output:
                yield [row[x] for x in schema.names]

        return IterableDataFrame(get_all(), schema)

    @no_type_check
    def count(self, df: EmptyAwareIterable[Dict[str, Any]]) -> int:
        return sum(1 for _ in df)


@fugue_annotated_param(pd.DataFrame, "p")
class _PandasParam(LocalDataFrameParam):
    @no_type_check
    def to_input_data(self, df: DataFrame, ctx: Any) -> pd.DataFrame:
        return df.as_pandas()

    @no_type_check
    def to_output_df(self, output: pd.DataFrame, schema: Any, ctx: Any) -> DataFrame:
        _schema: Optional[Schema] = None if schema is None else Schema(schema)
        if _schema is not None and _schema.names != list(output.columns):
            output = output[_schema.names]
        return PandasDataFrame(output, schema)

    @no_type_check
    def count(self, df: pd.DataFrame) -> int:
        return df.shape[0]

    def format_hint(self) -> Optional[str]:
        return "pandas"


@fugue_annotated_param(
    Iterable[pd.DataFrame],
    matcher=lambda x: x == Iterable[pd.DataFrame] or x == Iterator[pd.DataFrame],
)
class _IterablePandasParam(LocalDataFrameParam):
    @no_type_check
    def to_input_data(self, df: DataFrame, ctx: Any) -> Iterable[pd.DataFrame]:
        if not isinstance(df, LocalDataFrameIterableDataFrame):
            yield df.as_pandas()
        else:
            for sub in df.native:
                yield sub.as_pandas()

    @no_type_check
    def to_output_df(
        self, output: Iterable[pd.DataFrame], schema: Any, ctx: Any
    ) -> DataFrame:
        def dfs():
            _schema: Optional[Schema] = None if schema is None else Schema(schema)
            has_return = False
            for df in output:
                if _schema is not None and _schema.names != list(df.columns):
                    df = df[_schema.names]
                yield PandasDataFrame(df, _schema)
                has_return = True
            if not has_return and _schema is not None:
                yield PandasDataFrame(schema=_schema)

        return IterablePandasDataFrame(dfs())

    @no_type_check
    def count(self, df: Iterable[pd.DataFrame]) -> int:
        return sum(_.shape[0] for _ in df)

    def format_hint(self) -> Optional[str]:
        return "pandas"


@fugue_annotated_param(pa.Table)
class _PyArrowTableParam(LocalDataFrameParam):
    def to_input_data(self, df: DataFrame, ctx: Any) -> Any:
        return df.as_arrow()

    def to_output_df(self, output: Any, schema: Any, ctx: Any) -> DataFrame:
        assert isinstance(output, pa.Table)
        adf: DataFrame = ArrowDataFrame(output)
        if schema is not None:
            _schema = Schema(schema)
            if adf.schema != _schema:
                adf = adf[_schema.names].alter_columns(_schema)
        return adf

    def count(self, df: Any) -> int:  # pragma: no cover
        return df.count()

    def format_hint(self) -> Optional[str]:
        return "pyarrow"


@fugue_annotated_param(
    Iterable[pa.Table],
    matcher=lambda x: x == Iterable[pa.Table] or x == Iterator[pa.Table],
)
class _IterableArrowParam(LocalDataFrameParam):
    @no_type_check
    def to_input_data(self, df: DataFrame, ctx: Any) -> Iterable[pa.Table]:
        if not isinstance(df, LocalDataFrameIterableDataFrame):
            yield df.as_arrow()
        else:
            for sub in df.native:
                yield sub.as_arrow()

    @no_type_check
    def to_output_df(
        self, output: Iterable[pa.Table], schema: Any, ctx: Any
    ) -> DataFrame:
        def dfs():
            _schema: Optional[Schema] = None if schema is None else Schema(schema)
            has_return = False
            for df in output:
                adf: DataFrame = ArrowDataFrame(df)
                if _schema is not None and adf.schema != _schema:
                    adf = adf[_schema.names].alter_columns(_schema)
                yield adf
                has_return = True
            if not has_return and _schema is not None:
                yield ArrowDataFrame(schema=_schema)

        return IterableArrowDataFrame(dfs())

    @no_type_check
    def count(self, df: Iterable[pa.Table]) -> int:
        return sum(_.shape[0] for _ in df)

    def format_hint(self) -> Optional[str]:
        return "pyarrow"


@fugue_annotated_param(DataFrames, "c")
class _DataFramesParam(AnnotatedParam):
    pass



class FugueError(Exception):
    """Fugue exceptions"""


class FugueBug(FugueError):
    """Fugue internal bug"""


class FugueInvalidOperation(FugueError):
    """Invalid operation on the Fugue framework"""


class FuguePluginsRegistrationError(FugueError):
    """Fugue plugins registration error"""


class FugueDataFrameError(FugueError):
    """Fugue dataframe related error"""


class FugueDataFrameInitError(FugueDataFrameError):
    """Fugue dataframe initialization error"""


class FugueDatasetEmptyError(FugueDataFrameError):
    """Fugue dataframe is empty"""


class FugueDataFrameOperationError(FugueDataFrameError):
    """Fugue dataframe invalid operation"""


class FugueWorkflowError(FugueError):
    """Fugue workflow exceptions"""


class FugueWorkflowCompileError(FugueWorkflowError):
    """Fugue workflow compile time error"""


class FugueWorkflowCompileValidationError(FugueWorkflowCompileError):
    """Fugue workflow compile time validation error"""


class FugueInterfacelessError(FugueWorkflowCompileError):
    """Fugue interfaceless exceptions"""


class FugueWorkflowRuntimeError(FugueWorkflowError):
    """Fugue workflow compile time error"""


class FugueWorkflowRuntimeValidationError(FugueWorkflowRuntimeError):
    """Fugue workflow runtime validation error"""


class FugueSQLError(FugueWorkflowCompileError):
    """Fugue SQL error"""


class FugueSQLSyntaxError(FugueSQLError):
    """Fugue SQL syntax error"""


class FugueSQLRuntimeError(FugueWorkflowRuntimeError):
    """Fugue SQL runtime error"""



# flake8: noqa
from triad.collections import Schema

from fugue.api import out_transform, transform
from fugue.bag.array_bag import ArrayBag
from fugue.bag.bag import Bag, BagDisplay
from fugue.collections.partition import PartitionCursor, PartitionSpec
from fugue.collections.sql import StructuredRawSQL, TempTableName
from fugue.collections.yielded import PhysicalYielded, Yielded
from fugue.constants import register_global_conf
from fugue.dataframe.array_dataframe import ArrayDataFrame
from fugue.dataframe.arrow_dataframe import ArrowDataFrame
from fugue.dataframe.dataframe import (
    AnyDataFrame,
    DataFrame,
    DataFrameDisplay,
    LocalBoundedDataFrame,
    LocalDataFrame,
)
from fugue.dataframe.dataframe_iterable_dataframe import (
    IterableArrowDataFrame,
    IterablePandasDataFrame,
    LocalDataFrameIterableDataFrame,
)
from fugue.dataframe.dataframes import DataFrames
from fugue.dataframe.iterable_dataframe import IterableDataFrame
from fugue.dataframe.pandas_dataframe import PandasDataFrame
from fugue.dataset import (
    AnyDataset,
    Dataset,
    DatasetDisplay,
    as_fugue_dataset,
    get_dataset_display,
)
from fugue.execution.execution_engine import (
    AnyExecutionEngine,
    EngineFacet,
    ExecutionEngine,
    MapEngine,
    SQLEngine,
)
from fugue.execution.factory import (
    is_pandas_or,
    make_execution_engine,
    make_sql_engine,
    register_default_execution_engine,
    register_default_sql_engine,
    register_execution_engine,
    register_sql_engine,
)
from fugue.execution.native_execution_engine import (
    NativeExecutionEngine,
    PandasMapEngine,
    QPDPandasEngine,
)
from fugue.extensions.creator import Creator, creator, register_creator
from fugue.extensions.outputter import Outputter, outputter, register_outputter
from fugue.extensions.processor import Processor, processor, register_processor
from fugue.extensions.transformer import (
    CoTransformer,
    OutputCoTransformer,
    OutputTransformer,
    Transformer,
    cotransformer,
    output_cotransformer,
    output_transformer,
    register_output_transformer,
    register_transformer,
    transformer,
)
from fugue.registry import _register
from fugue.rpc import (
    EmptyRPCHandler,
    RPCClient,
    RPCFunc,
    RPCHandler,
    RPCServer,
    make_rpc_server,
    to_rpc_handler,
)
from fugue.sql.api import fugue_sql_flow as fsql
from fugue.sql.workflow import FugueSQLWorkflow
from fugue.workflow._workflow_context import FugueWorkflowContext
from fugue.workflow.module import module
from fugue.workflow.workflow import FugueWorkflow, WorkflowDataFrame, WorkflowDataFrames
from fugue_version import __version__

from .dev import *

_register()



from abc import abstractmethod
from typing import Any, Callable

import ibis

from fugue import AnyDataFrame, DataFrame, DataFrames, EngineFacet, ExecutionEngine
from fugue._utils.registry import fugue_plugin

from .._compat import IbisTable


@fugue_plugin
def parse_ibis_engine(obj: Any, engine: ExecutionEngine) -> "IbisEngine":
    if isinstance(obj, IbisEngine):
        return obj
    raise NotImplementedError(
        f"Ibis execution engine can't be parsed from {obj}."
        " You may need to register a parser for it."
    )


class IbisEngine(EngineFacet):
    """The abstract base class for different ibis execution implementations.

    :param execution_engine: the execution engine this ibis engine will run on
    """

    @property
    def is_distributed(self) -> bool:  # pragma: no cover
        return self.execution_engine.is_distributed

    def to_df(self, df: AnyDataFrame, schema: Any = None) -> DataFrame:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def select(
        self, dfs: DataFrames, ibis_func: Callable[[ibis.BaseBackend], IbisTable]
    ) -> DataFrame:  # pragma: no cover
        """Execute the ibis select expression.

        :param dfs: a collection of dataframes that must have keys
        :param ibis_func: the ibis compute function
        :return: result of the ibis function

        .. note::

            This interface is experimental, so it is subjected to change.
        """
        raise NotImplementedError



# flake8: noqa
# pylint: disable-all
from .dataframe.api import (
    alter_columns,
    as_array,
    as_array_iterable,
    as_arrow,
    as_dict_iterable,
    as_dicts,
    as_fugue_df,
    as_pandas,
    drop_columns,
    get_column_names,
    get_native_as_df,
    get_schema,
    head,
    is_df,
    normalize_column_names,
    peek_array,
    peek_dict,
    rename,
    select_columns,
)
from .dataset.api import (
    as_fugue_dataset,
    as_local,
    as_local_bounded,
    count,
    get_num_partitions,
    is_bounded,
    is_empty,
    is_local,
    show,
)
from .execution.api import (
    aggregate,
    anti_join,
    as_fugue_engine_df,
    assign,
    broadcast,
    clear_global_engine,
    cross_join,
    distinct,
    dropna,
    engine_context,
    fillna,
    filter,
    full_outer_join,
    get_context_engine,
    get_current_conf,
    get_current_parallelism,
    inner_join,
    intersect,
    join,
    left_outer_join,
    load,
    persist,
    repartition,
    right_outer_join,
    run_engine_function,
    sample,
    save,
    select,
    semi_join,
    set_global_engine,
    subtract,
    take,
    union,
)
from .sql.api import fugue_sql, fugue_sql_flow
from .workflow.api import out_transform, raw_sql, transform



# pylint: disable=W0611,W0613
import html
import json
from typing import Any, Dict, List, Optional

from IPython import get_ipython
from IPython.core.magic import Magics, cell_magic, magics_class, needs_local_scope
from IPython.display import HTML, display
from triad import ParamDict
from triad.utils.convert import to_instance
from triad.utils.pyarrow import _field_to_expression

from fugue import DataFrame, DataFrameDisplay, ExecutionEngine
from fugue import fsql as fugue_sql
from fugue import get_dataset_display, make_execution_engine
from fugue.dataframe import YieldedDataFrame
from fugue.exceptions import FugueSQLSyntaxError


class NotebookSetup(object):
    """Jupyter notebook environment customization template."""

    def get_pre_conf(self) -> Dict[str, Any]:
        """The default config for all registered execution engine"""
        return {}

    def get_post_conf(self) -> Dict[str, Any]:
        """The enforced config for all registered execution engine.
        Users should not set these configs manually, if they set, the values
        must match this dict, otherwise, exceptions will be thrown
        """
        return {}


@magics_class
class _FugueSQLMagics(Magics):
    """Fugue SQL Magics"""

    def __init__(
        self,
        shell: Any,
        pre_conf: Dict[str, Any],
        post_conf: Dict[str, Any],
        fsql_ignore_case: bool = False,
    ):
        # You must call the parent constructor
        super().__init__(shell)
        self._pre_conf = pre_conf
        self._post_conf = post_conf
        self._fsql_ignore_case = fsql_ignore_case

    @needs_local_scope
    @cell_magic("fsql")
    def fsql(self, line: str, cell: str, local_ns: Any = None) -> None:
        try:
            dag = fugue_sql(
                "\n" + cell, local_ns, fsql_ignore_case=self._fsql_ignore_case
            )
        except FugueSQLSyntaxError as ex:
            raise FugueSQLSyntaxError(str(ex)).with_traceback(None) from None
        dag.run(self.get_engine(line, {} if local_ns is None else local_ns))
        for k, v in dag.yields.items():
            if isinstance(v, YieldedDataFrame):
                local_ns[k] = v.result  # type: ignore
            else:
                local_ns[k] = v  # type: ignore

    def get_engine(self, line: str, lc: Dict[str, Any]) -> ExecutionEngine:
        line = line.strip()
        p = line.find("{")
        if p >= 0:
            engine = line[:p].strip()
            conf = json.loads(line[p:])
        else:
            parts = line.split(" ", 1)
            engine = parts[0]
            conf = ParamDict(None if len(parts) == 1 else lc[parts[1]])
        cf = dict(self._pre_conf)
        cf.update(conf)
        for k, v in self._post_conf.items():
            if k in cf and cf[k] != v:
                raise ValueError(
                    f"{k} must be {v}, but you set to {cf[k]}, you may unset it"
                )
            cf[k] = v
        if "+" in engine:
            return make_execution_engine(tuple(engine.split("+", 1)), cf)
        return make_execution_engine(engine, cf)


class JupyterDataFrameDisplay(DataFrameDisplay):
    def show(
        self, n: int = 10, with_count: bool = False, title: Optional[str] = None
    ) -> None:
        components: List[Any] = []
        if title is not None:
            components.append(HTML(f"<h3>{html.escape(title)}</h3>"))
        if with_count:
            count = self.df.count()
        else:
            count = -1
        components.append(HTML(self._generate_df_html(n)))
        if count >= 0:
            components.append(HTML(f"<strong>total count: {count}</strong>"))
        display(*components)

    def repr_html(self) -> str:
        return self._generate_df_html(10)

    def _generate_df_html(self, n: int) -> str:
        res: List[str] = []
        pdf = self.df.head(n).as_pandas()
        cols = [_field_to_expression(f) for f in self.df.schema.fields]
        pdf.columns = cols
        res.append(pdf._repr_html_())
        schema = type(self.df).__name__ + ": " + str(self.df.schema)
        res.append('<font size="-1">' + html.escape(schema) + "</font>")
        return "\n".join(res)


@get_dataset_display.candidate(
    lambda ds: get_ipython() is not None and isinstance(ds, DataFrame), priority=3.0
)
def _get_jupyter_dataframe_display(ds: DataFrame):
    return JupyterDataFrameDisplay(ds)


def _setup_fugue_notebook(
    ipython: Any, setup_obj: Any, fsql_ignore_case: bool = False
) -> None:
    s = NotebookSetup() if setup_obj is None else to_instance(setup_obj, NotebookSetup)
    magics = _FugueSQLMagics(
        ipython,
        dict(s.get_pre_conf()),
        dict(s.get_post_conf()),
        fsql_ignore_case=fsql_ignore_case,
    )
    ipython.register_magics(magics)
